# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
import json
import os

# generate build configuration files for use with https://github.com/enzienaudio/courtesan
# Example arguments:
#   macos_x64_args=["-project", "Hv_test_WwiseSourcePlugin.xcodeproj", "-arch", "x86_64", "-alltargets"]
#   android_armv7a_args=["APP_ABI=armeabi-v7a", "-j"]
#   win_x64_args=["/property:Configuration=Release", "/property:Platform=x64",
#                 "/t:Rebuild", "Hv_{0}_Unity.sln".format(patch_name), "/m"]


def generate_json(out_dir, android_armv7a_args=None, ios_armv7a_args=None,
                  linux_armv7a_args=None, linux_x64_args=None, macos_x64_args=None,
                  win_x64_args=None, win_x86_args=None):
    build_json = defaultdict(dict)

    if android_armv7a_args:
        build_json["android"]["armv7a"] = {
            "args": [android_armv7a_args],
            "projectDir": ["android", "jni"],
            "binaryDir": ["android", "libs", "armeabi-v7a"]
        }
    if ios_armv7a_args:
        build_json["ios"]["armv7a"] = {
            "args": [ios_armv7a_args],
            "projectDir": ["xcode"],
            "binaryDir": ["build", "ios", "armv7s", "Release"]
        }
    if linux_armv7a_args:
        build_json["linux"]["armv7a"] = {
            "args": [linux_armv7a_args],
            "projectDir": ["linux"],
            "binaryDir": ["build", "linux", "armv7a", "release"]
        }
    if linux_x64_args:
        build_json["linux"]["x64"] = {
            "args": [linux_x64_args],
            "projectDir": ["linux"],
            "binaryDir": ["build", "linux", "x64", "release"]
        }
    if macos_x64_args:
        build_json["macos"]["x64"] = {
            "args": [macos_x64_args],
            "projectDir": ["xcode"],
            "binaryDir": ["build", "macos", "x86_64", "Release"]
        }
    if win_x64_args:
        build_json["win"]["x64"] = {
            "args": [win_x64_args],
            "projectDir": ["vs2015"],
            "binaryDir": ["build", "win", "x64", "Release"]
        }
    if win_x86_args:
        build_json["win"]["x86"] = {
            "args": [win_x86_args],
            "projectDir": ["vs2015"],
            "binaryDir": ["build", "win", "x86", "Release"]
        }

    with open(os.path.join(out_dir, "build.json"), "w") as f:
        json.dump(build_json, f)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import decimal
import json
import os
import random
from struct import unpack, pack
import string

from hvcc.core.hv2ir.HeavyException import HeavyException


class HeavyLangObject:
    """ This is the base Heavy object class.
    """

    __RANDOM = random.Random()
    __ID_CHARS = string.ascii_letters + string.digits

    # load the Heavy object definitions
    with open(os.path.join(os.path.dirname(__file__), "../json/heavy.lang.json"), "r") as f:
        _HEAVY_LANG_DICT = json.load(f)

    def __init__(self, obj_type, args=None, graph=None, num_inlets=-1, num_outlets=-1, annotations=None):
        # set the object type
        self.type = obj_type

        # generate a unique id for this object
        self.id = "".join(HeavyLangObject.__RANDOM.choice(HeavyLangObject.__ID_CHARS) for _ in range(8))

        # assign the parent graph
        self.graph = graph

        # set local arguments
        self.args = args or {}

        # set local annotations
        self.annotations = annotations or {}

        # a list of locally generated warnings and errors (notifications)
        self.warnings = []
        self.errors = []

        # resolve arguments and fill in missing defaults for HeavyLang objects
        self.__resolve_default_lang_args()

        # the list of connections at each inlet
        num_inlets = num_inlets if num_inlets >= 0 else len(self._obj_desc["inlets"])
        self.inlet_connections = [[] for _ in range(num_inlets)]

        # the list of connections at each outlet
        num_outlets = num_outlets if num_outlets >= 0 else len(self._obj_desc["outlets"])
        self.outlet_connections = [[] for _ in range(num_outlets)]

    @property
    def scope(self):
        """ Returns the scope of this object, private by default.
            Scope may be public, protected, private.
        """
        return self.annotations.get("scope", "private")

    @property
    def static(self):
        """ Returns true of this object is marked as static. False by default.
        """
        return self.annotations.get("static", False)

    @property
    def const(self):
        """ Returns true of this object is marked as constant. False by default.
        """
        return self.annotations.get("const", False)

    @property
    def name(self):
        """ Returns the name of this object. Returns None if there is no name.
        """
        return self.args.get("name", None)

    @property
    def _obj_desc(self):
        """ Returns the HeavyLang object description.
        """
        return HeavyLangObject._HEAVY_LANG_DICT[self.type]

    def inlet_connection_type(self, index):
        return self._obj_desc["inlets"][index]

    def outlet_connection_type(self, index):
        return self._obj_desc["outlets"][index]

    def name_for_arg(self, index=0):
        """ Returns the name of the argument at the given index.
        """
        return self._obj_desc["args"][index]["name"]

    def add_warning(self, warning):
        """ Add a warning to this object.
        """
        self.warnings.append({"message": warning})

    def add_error(self, error):
        """ Add an error to this object and raise an exception.
        """
        self.errors.append({"message": error})
        raise HeavyException(error)

    def get_notices(self):
        """ Returns a dictionary of all warnings and errors at this object.
        """
        return {
            "warnings": [{"message": "{0}: {1}".format(self, n["message"])} for n in self.warnings],
            "errors": [{"message": "{0}: {1}".format(self, n["message"])} for n in self.errors],
        }

    @classmethod
    def force_arg_type(clazz, value, value_type, graph=None):
        """ Attempts to convert a value to a given value type. Raises an Exception otherwise.
            If the value_type is unknown and a graph is provided, a warning will be registered.
        """
        if value_type == "float":
            try:
                return float(value)
            except Exception:
                raise Exception(f"Cannot convert argument \"{value}\" into float.")
        elif value_type == "int":
            return int(decimal.Decimal(value))
        elif value_type == "string":
            return str(value) if value is not None else None
        elif value_type == "boolean":
            if isinstance(value, str):
                return value.strip().lower() not in ["false", "f", "0"]
            else:
                return bool(value)
        elif value_type == "floatarray":
            if isinstance(value, list):
                return [float(v) for v in value]
            if isinstance(value, str):
                return [float(v) for v in value.split()]
            else:
                raise HeavyException(f"Cannot convert value to type floatarray: {value}")
        elif value_type == "intarray":
            if isinstance(value, list):
                return [int(v) for v in value]
            if isinstance(value, str):
                return [int(v) for v in value.split()]
            else:
                raise HeavyException(f"Cannot convert value to type intarray: {value}")
        elif value_type == "stringarray":
            if isinstance(value, list):
                return [str(v) for v in value]
            if isinstance(value, str):
                return [str(v) for v in value.split()]
            else:
                raise HeavyException(f"Cannot convert value to type stringarray: {value}")
        else:
            # NOTE(mhroth): if value_type is not a known type or None, that is
            # not necessarily an error. It may simply be that the value should
            # not be resolved to anything other than what it already is.
            # This happens most often with message objects.
            # if graph is not None:
            #     graph.add_warning("Unknown value type \"{0}\" for value: {1}".format(value_type, value))
            return value

    def __resolve_default_lang_args(self):
        """ Resolves missing default arguments. Also checks to make sure that all
            required arguments are present. Does nothing if the object is IR.
        """
        if self.type in HeavyLangObject._HEAVY_LANG_DICT:
            for arg in self._obj_desc["args"]:
                if arg["name"] not in self.args:
                    # if a defined argument is not in the argument dictionary
                    if not arg["required"]:
                        # if the argument is not required, use the default
                        self.args[arg["name"]] = arg["default"]
                    else:
                        self.add_error("Required argument \"{0}\" not present for object {1}.".format(
                            arg["name"],
                            self))
                else:
                    # enforce argument types
                    self.args[arg["name"]] = HeavyLangObject.force_arg_type(
                        self.args[arg["name"]],
                        arg["value_type"],
                        self.graph)

    @property
    def num_inlets(self):
        return len(self.inlet_connections)

    @property
    def num_outlets(self):
        return len(self.outlet_connections)

    def add_connection(self, c):
        """ Add a connection to this object.
        """
        try:
            if c.to_object is self:
                self.inlet_connections[c.inlet_index].append(c)
            elif c.from_object is self:
                self.outlet_connections[c.outlet_index].append(c)
            else:
                raise HeavyException("Connection {0} does not connect to this object {1}.".format(c, self))
        except Exception:
            raise HeavyException("Connection {0} connects to out-of-range let.".format(c))

    def remove_connection(self, c):
        """ Remove a connection to this object.
        """
        if c.to_object is self:
            self.inlet_connections[c.inlet_index].remove(c)
        elif c.from_object is self:
            self.outlet_connections[c.outlet_index].remove(c)
        else:
            raise HeavyException("Connection {0} does not connect to this object {1}.".format(c, self))

    def replace_connection(self, c, n_list):
        """ Replaces connection c with connection list n_list, maintaining connection order
        """
        if c.from_object is self:
            cc = self.outlet_connections[c.outlet_index]
            # NOTE(mhroth): this will throw an exception if c does not exist
            # if a heavy object connects to itself, such as through a [t a]
            # or directly, there may be an error here,
            i = cc.index(c)
            self.outlet_connections[c.outlet_index] = cc[0:i] + n_list + cc[i + 1:]
        elif c.to_object is self:
            # connection order doesn't matter at the inlet
            self.inlet_connections[c.inlet_index].remove(c)
            self.inlet_connections[c.inlet_index].extend(n_list)
        else:
            raise HeavyException("Connections must have a common endpoint: {0} / {1}".format(c, n_list))

    def get_connection_move_list(self, o, connection_type_filter="-~>"):
        """ Create a list of commands to move all connections from this object
            to the given object o.
        """
        m = []
        for c in [c for cc in self.inlet_connections for c in cc]:
            m.append((c, [c.copy(to_object=o)]))
        for c in [c for cc in self.outlet_connections for c in cc]:
            m.append((c, [c.copy(from_object=o)]))
        return m

    def _get_connection_format(self, connections_list):
        fmt = []
        for cc in connections_list:
            s = {c.type for c in cc}
            if len(s) == 0:
                fmt.append("_")
            elif len(s) == 1:
                if "-->" in s:
                    fmt.append("c")
                elif "~f>" in s:
                    fmt.append("f")
                elif "~i>" in s:
                    fmt.append("i")
                else:
                    raise Exception("Unknown connection type in set {0} in file {1}.".format(
                        cc,
                        cc[0].from_object.graph.file))
            elif s in [{"~f>", "-->"}, {"~i>", "-->"}]:
                fmt.append("m")
            else:
                fmt.append("m")
        return "".join(fmt)

    def has_inlet_connection_format(self, fmts=[]):
        """ Returns true if the object has given format at its inlets.
        """
        fmts = fmts if isinstance(fmts, list) else [fmts]
        return self._get_connection_format(self.inlet_connections) in fmts

    def has_outlet_connection_format(self, fmts=[]):
        """ Returns true if the object has given format at its outlets.
            Take either litteral or list as input.
        """
        fmts = fmts if isinstance(fmts, list) else [fmts]
        return self._get_connection_format(self.outlet_connections) in fmts

    def is_leaf(self):
        """ Returns True if this object is a leaf in the graph. False otherwise.
        """
        return all(len(c) == 0 for c in self.outlet_connections)

    def is_root(self):
        """ Returns True if this object is a root in the graph. False otherwise.
        """
        return all(len(c) == 0 for c in self.inlet_connections)

    def _resolved_outlet_type(self, outlet_index=0):
        """ Returns the connection type expected at the given outlet.
            The result may be influenced by the state of the input connections.
        """
        # get the defined connection type
        connection_type = self._obj_desc["outlets"][outlet_index]["connectionType"]
        if connection_type == "-~>":
            # if the connection type is defined as mixed,
            # use the default approach to resolve it
            connection_type_set = {c.type for c in self.inlet_connections[0]}
            if len(connection_type_set) == 0:
                return "-->"
            elif len(connection_type_set) == 1:
                return list(connection_type_set)[0]
            elif len(connection_type_set) == 2:
                if {"-->", "~f>"} == connection_type_set:
                    return "~f>"
                elif {"-->", "~i>"} == connection_type_set:
                    return "~i>"
            # if the connection type cannot be resolved to a well-defined type
            self.graph.add_error("Connection type cannot be resolved."
                                 f"Unknown inlet connection type configuration: {connection_type_set}")
        else:
            return connection_type

    def _resolve_connection_types(self, obj_stack):
        """ Resolves the type of all connections before reduction to IR object types.
            If connections incident on an object are incompatible, they are either
            resolved, potentially by inserting conversion objects, or pruned.
            If a viable resolution cannot be found, an exception may be raised.
        """
        if self in obj_stack:
            return
        else:
            obj_stack.add(self)

        # for all outgoing connections
        for c in [c for cc in self.outlet_connections for c in cc]:
            if c.is_mixed:
                connection_type = self._resolved_outlet_type(outlet_index=c.outlet_index)
                if connection_type == "-~>":
                    self.graph.add_error("Cannot resolve connection type from -~>.")
                else:
                    self.graph.update_connection(c, [c.copy(type=connection_type)])

            c.to_object._resolve_connection_types(obj_stack)  # turtle all the way down

    @classmethod
    def get_hash(clazz, x):
        """ Compute the message element hash used by msg_getHash(). Returns a 32-bit integer.
        """
        if isinstance(x, float) or isinstance(x, int):
            # interpret the float bytes as an unsigned integer
            return unpack("@I", pack("@f", float(x)))[0]
        elif x == "bang":
            return 0xFFFFFFFF
        elif isinstance(x, str):
            # this hash is based MurmurHash2
            # http://en.wikipedia.org/wiki/MurmurHash
            # https://sites.google.com/site/murmurhash/
            x = str(x)
            m = 0x5bd1e995
            r = 24
            h = len(x)
            i = 0
            while i < len(x) & ~0x3:
                k = unpack("@I", bytes(x[i:i + 4], "utf-8"))[0]
                k = (k * m) & 0xFFFFFFFF
                k ^= k >> r
                k = (k * m) & 0xFFFFFFFF
                h = (h * m) & 0xFFFFFFFF
                h ^= k
                i += 4

            n = len(x) & 0x3
            x = x[i:i + n]
            if n >= 3:
                h ^= (ord(x[2]) << 16) & 0xFFFFFFFF
            if n >= 2:
                h ^= (ord(x[1]) << 8) & 0xFFFFFFFF
            if n >= 1:
                h ^= ord(x[0])
                h = (h * m) & 0xFFFFFFFF

            h ^= h >> 13
            h = (h * m) & 0xFFFFFFFF
            h ^= h >> 15
            return h
        else:
            raise Exception("Message element hashes can only be computed for float and string types.")

    def __repr__(self):
        arg_str = " ".join(["{0}:{1}".format(k, o) for (k, o) in self.args.iteritems()])
        return "{0} {{{1}}}".format(self.type, arg_str) if len(arg_str) > 0 else self.type



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject


class HLangAdc(HeavyLangObject):
    """ adc
    """

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(
            self,
            obj_type,
            args, graph,
            num_inlets=0,
            num_outlets=len(args[HeavyLangObject._HEAVY_LANG_DICT[obj_type]["args"][0]["name"]]),
            annotations=annotations)

    def _resolved_outlet_type(self, outlet_index=0):
        return "~f>"

    def reduce(self):
        objects = set()
        connections = []

        # reduce a HeavyLang adc to a number of individual HeavyIR __inlet objects
        for i, channel_index in enumerate(self.args["channels"]):
            if len(self.outlet_connections[i]) > 0:  # if there are any connections to this inlet
                x = HeavyIrObject("__inlet", args={
                    "index": 127 + channel_index
                })
                x.outlet_buffers[0] = ("input", channel_index - 1)  # channel indicies are one-indexed
                objects.add(x)

                for c in self.outlet_connections[i]:
                    connections.append((c, [c.copy(from_object=x, outlet_index=0)]))

        return (objects, connections)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalRPole(HeavyObject):
    """Handles the __rpole~f object.
    """

    c_struct = "SignalRpole"
    preamble = "sRPole"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalRPole.h", "HvSignalDel1.h", "HvMath.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {
            "HvSignalRPole.h", "HvSignalRPole.c",
            "HvSignalDel1.h", "HvSignalDel1.c",
            "HvMath.h"
        }

    @classmethod
    def get_C_def(clazz, obj_type, obj_id):
        return ["SignalRPole sRPole_{0};".format(obj_id)]

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return ["sRPole_init(&sRPole_{0});".format(obj_id)]

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, objects):
        return [
            "__hv_rpole_f(&sRPole_{0}, VIf({1}), VIf({2}), VOf({3}));".format(
                process_dict["id"],
                HeavyObject._c_buffer(process_dict["inputBuffers"][0]),
                HeavyObject._c_buffer(process_dict["inputBuffers"][1]),
                HeavyObject._c_buffer(process_dict["outputBuffers"][0]))]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalCPole(HeavyObject):
    """Handles the __cpole~f object.
    """

    c_struct = "SignalCPole"
    preamble = "sCPole"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalCPole.h", "HvSignalDel1.h", "HvMath.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {
            "HvSignalCPole.h", "HvSignalCPole.c",
            "HvSignalDel1.h", "HvSignalDel1.c",
            "HvMath.h"
        }

    @classmethod
    def get_C_def(clazz, obj_type, obj_id):
        return ["SignalCPole sCPole_{0};".format(obj_id)]

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return ["sCPole_init(&sCPole_{0});".format(obj_id)]

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, objects):
        return [
            "__hv_cpole_f(&sCPole_{0}, VIf({1}), VIf({2}), VIf({3}), VIf({4}), VOf({5}), VOf({6}));".format(
                process_dict["id"],
                HeavyObject._c_buffer(process_dict["inputBuffers"][0]),
                HeavyObject._c_buffer(process_dict["inputBuffers"][1]),
                HeavyObject._c_buffer(process_dict["inputBuffers"][2]),
                HeavyObject._c_buffer(process_dict["inputBuffers"][3]),
                HeavyObject._c_buffer(process_dict["outputBuffers"][0]),
                HeavyObject._c_buffer(process_dict["outputBuffers"][1]))]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import time

from .HeavyException import HeavyException
from .HeavyParser import HeavyParser


class hv2ir:

    @classmethod
    def compile(clazz, hv_file, ir_file, patch_name=None, verbose=False):
        """ Compiles a HeavyLang file into a HeavyIR file.
            Returns a tuple of compile time in seconds, a notification dictionary,
            and a heavy object counter.
        """

        # keep track of the total compile time
        tick = time.time()

        hv_file = os.path.abspath(os.path.expanduser(hv_file))
        ir_file = os.path.abspath(os.path.expanduser(ir_file))

        try:
            # parse heavy file
            hv_graph = HeavyParser.graph_from_file(hv_file=hv_file, xname=patch_name)
        except HeavyException as e:
            return {
                "stage": "hv2ir",
                "compile_time": time.time() - tick,
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "errors": [{"message": e.message}],
                    "warnings": []
                },
                "obj_counter": None,
                "in_file": os.path.basename(hv_file),
                "in_dir": os.path.dirname(hv_file),
                "out_file": os.path.basename(ir_file),
                "out_dir": os.path.dirname(ir_file)
            }

        try:
            # get a counter of all heavy objects
            hv_counter = hv_graph.get_object_counter(recursive=True)

            # prepare the graph for exporting
            hv_graph.prepare()

            # ensure that the output directory exists
            if not os.path.exists(os.path.dirname(ir_file)):
                os.makedirs(os.path.dirname(ir_file))

            # generate Heavy.IR
            ir = hv_graph.to_ir()
        except HeavyException as e:
            return {
                "stage": "hv2ir",
                "compile_time": time.time() - tick,
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "errors": [{"message": e.message}],
                    "warnings": []
                },
                "obj_counter": hv_counter,
                "in_file": os.path.basename(hv_file),
                "in_dir": os.path.dirname(hv_file),
                "out_file": os.path.basename(ir_file),
                "out_dir": os.path.dirname(ir_file)
            }

        # write the hv.ir file
        with open(ir_file, "w") as f:
            if verbose:
                json.dump(
                    ir,
                    f,
                    sort_keys=True,
                    indent=2,
                    separators=(",", ": "))
            else:
                json.dump(ir, f)

        if verbose:
            if len(ir["signal"]["processOrder"]) > 0:
                print("")
                print("=== Signal Order ===")
                for so in ir["signal"]["processOrder"]:
                    o = ir["objects"][so["id"]]
                    if len(o["args"]) > 0:
                        print("{0} {{{1}}}".format(
                            o["type"],
                            " ".join(["{0}:{1}".format(k, v) for k, v in o["args"].items()])))
                    else:
                        print(o["type"])

        return {
            "stage": "hv2ir",
            "compile_time": time.time() - tick,  # record the total compile time
            "notifs": hv_graph.get_notices(),
            "obj_counter": hv_counter,
            "in_file": os.path.basename(hv_file),
            "in_dir": os.path.dirname(hv_file),
            "out_file": os.path.basename(ir_file),
            "out_dir": os.path.dirname(ir_file),
            "ir": ir
        }


def main():
    parser = argparse.ArgumentParser(
        description="A C-language compiler for the Heavy audio programming language.")
    parser.add_argument(
        "hv_path",
        help="The patch to the top-level patch to compile.")
    parser.add_argument(
        "--hv_ir_path",
        default="./heavy.hv.ir.json",
        help="The output path of the hv.ir.json file.")
    parser.add_argument(
        "--name",
        default="heavy",
        help="")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    d = hv2ir.compile(
        hv_file=args.hv_path,
        ir_file=args.hv_ir_path,
        patch_name=args.name,
        verbose=args.verbose)

    if args.verbose:
        print("Total hv2ir time: {0:.2f}ms".format(d["compile_time"] * 1000))


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject


class HLangBiquad(HeavyLangObject):
    """ Translates HeavyLang object biquad to HeavyIR biquad~.
    """

    # used to detect valid connection configuration formats
    __re_fmt_k = re.compile("[f_][c_]+")
    __re_fmt_f = re.compile("[f_]+")

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(self, "biquad", args, graph, num_inlets=6, num_outlets=1, annotations=annotations)

    def reduce(self):
        fmt = self._get_connection_format(self.inlet_connections)
        if HLangBiquad.__re_fmt_k.search(fmt):
            x = HeavyIrObject("__biquad_k~f", self.args)
            return {x}, self.get_connection_move_list(x)

        elif HLangBiquad.__re_fmt_f.search(fmt):
            x = HeavyIrObject("__biquad~f", self.args)
            return {x}, self.get_connection_move_list(x)

        else:
            raise HeavyException("Unsupported connection format to biquad: {0}".format(fmt))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlCast(HeavyObject):
    __OPERATION_DICT = {
        "__cast_b": "HV_CAST_BANG",
        "__cast_f": "HV_CAST_FLOAT",
        "__cast_s": "HV_CAST_SYMBOL"
    }

    c_struct = "ControlCast"
    preamble = "cCast"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvControlCast.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvControlCast.h", "HvControlCast.c"}

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [
            "cCast_onMessage(_c, {1}, 0, m, &cCast_{0}_sendMessage);".format(
                obj_id,
                ControlCast.__OPERATION_DICT[obj_type])
        ]



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

from .MaxObject import MaxObject


class MaxAdcObject(MaxObject):
    def __init__(self, obj_type, obj_args=None, obj_id=None, pos_x=0, pos_y=0):
        assert obj_type == "adc~"
        MaxObject.__init__(self, obj_type, obj_args, obj_id, pos_x, pos_y)

    def to_hv(self):
        return {
            "type": "adc",
            "args": {
                "channels": [1, 2] if len(self.obj_args) == 0 else [int(a) for a in self.obj_args]
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .Connection import Connection
from .HeavyLangObject import HeavyLangObject
from .HeavyIrObject import HeavyIrObject


class HLangBinop(HeavyLangObject):

    __HEAVY_DICT = {
        "+": ["__add", "__add_k", "__add~f", "__add~i"],
        "-": ["__sub", "__sub_k", "__sub~f", "__sub~i"],
        "*": ["__mul", "__mul_k", "__mul~f", "__mul~i"],
        "/": ["__div", "__div_k", "__div~f", "__div~i"],
        "max": ["__max", "__max_k", "__max~f", "__max~i"],
        "min": ["__min", "__min_k", "__min~f", "__min~i"],
        ">": ["__gt", "__gt_k", "__gt~f", "__gt~i"],
        ">=": ["__gte", "__gte_k", "__gte~f", "__gte~i"],
        "==": ["__eq", "__eq_k", "__eq~f", "__eq~i"],
        "!=": ["__neq", "__neq_k", "__neq~f", "__neq~i"],
        "<": ["__lt", "__lt_k", "__lt~f", "__lt~i"],
        "<=": ["__lte", "__lte_k", "__lte~f", "__lte~i"],
        "&": ["__and", "__and_k", "__and~f", "__and~i"],  # binary and
        "&&": ["__logand", "__logand_k"],  # logical or
        "&!": ["__andnot", "__andnot~f", "__andnot~i"],
        "|": ["__or", "__or_k", "__or~f", "__or~i"],  # binary or
        "||": ["__logor", "__logor_k"],  # logical or
        "pow": ["__pow", "__pow_k", "__pow~f", "__pow~i"],
        "atan2": ["__atan2", "__atan2_k", "__atan2~f"],
        "mod": ["__unimod", "__unimod_k"],
        "%": ["__bimod", "__bimod_k"],
        ">>": ["__shiftright", "__shiftright_k"],  # binary right shift
        "<<": ["__shiftleft", "__shiftleft_k"]  # binary left shift
    }

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(self, obj_type, args, graph, num_inlets=2, num_outlets=1, annotations=annotations)

    @classmethod
    def handles_type(clazz, obj_type):
        """Returns True if this class handles the given object type. False otherwise.
        """
        return obj_type in HLangBinop.__HEAVY_DICT

    def reduce(self):
        if self.has_inlet_connection_format("__") or \
                self.has_inlet_connection_format("_c") or \
                self.has_inlet_connection_format("_f") or \
                self.has_outlet_connection_format("_"):
            # binary operator objects must have a left inlet or outlet connection.
            return (set(), [])  # remove this object

        if self.has_inlet_connection_format("ff"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~f")][0]
            x = HeavyIrObject(ir_type)
            return ({x}, self.get_connection_move_list(x))

        elif self.has_inlet_connection_format("fc"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~f")][0]
            x = HeavyIrObject(ir_type)
            y = HeavyIrObject("__var~f", {"k": self.args[self.name_for_arg()]})
            z = HeavyIrObject("__varread~f", {"var_id": y.id})
            connections = [(None, [Connection(z, 0, x, 1, "~f>")])]
            for c in self.inlet_connections[0]:  # left inlet to ir_type
                connections.append((c, [c.copy(to_object=x)]))
            for c in self.inlet_connections[1]:  # right inlet to __var~f
                connections.append((c, [c.copy(to_object=y, inlet_index=0)]))
            for c in self.outlet_connections[0]:  # ir_type outlet
                connections.append((c, [c.copy(from_object=x)]))
            return ({x, y, z}, connections)

        elif self.has_inlet_connection_format("f_"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~f")][0]

            # handle identity-operations
            if (self.type == "+" and self.args["k"] == 0.0) or \
                    (self.type == "-" and self.args["k"] == 0.0) or \
                    (self.type == "*" and self.args["k"] == 1.0) or \
                    (self.type == "/" and self.args["k"] == 1.0):
                if len(self.inlet_connections[0]) == 0:
                    return (set(), [])  # remove this object
                elif len(self.inlet_connections[0]) == 1:
                    # there is only one connection, pass through the connection
                    # and remove this object
                    c = self.inlet_connections[0][0]
                    connections = [(c, [c.copy(to_object=v.to_object, inlet_index=v.inlet_index) for
                                        v in self.outlet_connections[0]])]
                    return (set(), connections)
                else:  # len(self.inlet_connections[0]) > 1
                    # there are multiple connections to the left inlet
                    # create a __add~f, move one connection to the left inlet
                    # and the remainder to the right inlet
                    x = HeavyIrObject("__add~f")
                    c = self.inlet_connections[0][0]
                    connections = [(c, [c.copy(to_object=x, inlet_index=0)])]
                    for c in self.inlet_connections[0][1:]:
                        connections.append((c, [c.copy(to_object=x, inlet_index=1)]))
                    for c in self.outlet_connections[0]:
                        connections.append((c, [c.copy(from_object=x, outlet_index=0)]))
                    return ({x}, connections)

            if self.type == "*":
                # self.args["k"] == 1.0 case handled above

                if self.args["k"] == 0.0:
                    # this object does nothing at all. Remove it.
                    return (set(), [])

                if self.args["k"] == -1.0:
                    x = HeavyIrObject("__neg~f")
                    return ({x}, self.get_connection_move_list(x))

            x = HeavyIrObject(ir_type)
            # add constant generator
            y = HeavyIrObject(
                "__var_k~f",
                args={"k": self.args[self.name_for_arg()]})
            connections = self.get_connection_move_list(x)
            connections.append((None, [Connection(y, 0, x, 1, "~f>")]))
            return ({x, y}, connections)

        elif self.has_inlet_connection_format("ii"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~i")][0]
            x = HeavyIrObject(ir_type)
            return ({x}, self.get_connection_move_list(x))

        elif self.has_inlet_connection_format("ic"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~i")][0]
            x = HeavyIrObject(ir_type)
            y = HeavyIrObject("__var~i", {"k": self.args[self.name_for_arg()]})
            z = HeavyIrObject("__varread~i", {"var_id": y.id})
            connections = [(None, [Connection(z, 0, x, 1, "~f>")])]
            for c in self.inlet_connections[0]:
                connections.append((c, [c.copy(to_object=x)]))
            for c in self.inlet_connections[1]:
                connections.append((c, [c.copy(to_object=y, inlet_index=0)]))
            for c in self.outlet_connections[0]:
                connections.append((c, [c.copy(from_object=x)]))
            return ({x, y, z}, connections)

        elif self.has_inlet_connection_format("i_"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~i")][0]
            x = HeavyIrObject(ir_type)
            # add constant generator
            y = HeavyIrObject(
                "__var_k~i",
                args={"k": self.args[self.name_for_arg()]})
            connections = [(None, [Connection(y, 0, x, 1, "~i>")])]
            for c in self.inlet_connections[0]:
                connections.append((c, [c.copy(to_object=x)]))
            for c in self.inlet_connections[1]:
                connections.append((c, [c.copy(to_object=y, inlet_index=0)]))
            for c in self.outlet_connections[0]:
                connections.append((c, [c.copy(from_object=x)]))
            return ({x, y}, connections)

        elif self.has_inlet_connection_format("cc"):
            ir_type = HLangBinop.__HEAVY_DICT[self.type][0]
            # standardise the operation's constant argument name to "k"
            args = {"k": self.args[self.name_for_arg()]}
            x = HeavyIrObject(ir_type, args)
            return ({x}, self.get_connection_move_list(x))

        elif self.has_inlet_connection_format("c_"):
            # if the right inlet is empty, generate a constant variation of the object
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("_k")][0]
            # standardise the operation's constant argument name to "k"
            args = {"k": self.args[self.name_for_arg()]}
            x = HeavyIrObject(ir_type, args)
            return ({x}, self.get_connection_move_list(x))

        elif self.has_inlet_connection_format("cf"):
            ir_type = [x for x in HLangBinop.__HEAVY_DICT[self.type] if x.endswith("~f")][0]
            x = HeavyIrObject(ir_type)
            v = HeavyIrObject("__var~f")
            vr = HeavyIrObject("__varread~f", args={"var_id": v.id})

            obj_set = set([x, v, vr])
            connections = [(None, [Connection(vr, 0, x, 0, "~f>")])]
            for c in self.inlet_connections[0]:
                connections.append((c, [c.copy(to_object=v, inlet_index=0)]))
            for c in self.inlet_connections[1]:
                connections.append((c, [c.copy(to_object=x, inlet_index=1)]))
            for c in self.outlet_connections[0]:
                connections.append((c, [c.copy(from_object=x, outlet_index=0)]))

            return (obj_set, connections)

        else:
            fmt = self._get_connection_format(self.inlet_connections)
            if "m" in fmt:
                self.add_error(
                    "A binary operator cannot have both a signal and control "
                    "connection to the same inlet: {0}".format(fmt))
            else:
                self.add_error(
                    "Unknown inlet configuration: {0}".format(fmt))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlUnop(HeavyObject):
    # a dictionary translating from the operation argument to the C case
    __OPERATION_DICT = {
        "__cos": "HV_UNOP_COS",
        "__acos": "HV_UNOP_ACOS",
        "__cosh": "HV_UNOP_COSH",
        "__acosh": "HV_UNOP_ACOSH",
        "__sin": "HV_UNOP_SIN",
        "__asin": "HV_UNOP_ASIN",
        "__sinh": "HV_UNOP_SINH",
        "__asinh": "HV_UNOP_ASINH",
        "__tan": "HV_UNOP_TAN",
        "__atan": "HV_UNOP_ATAN",
        "__tanh": "HV_UNOP_TANH",
        "__atanh": "HV_UNOP_ATANH",
        "__log": "HV_UNOP_LOG",
        "__log10": "HV_UNOP_LOG10",
        "__log2": "HV_UNOP_LOG2",
        "__exp": "HV_UNOP_EXP",
        "__abs": "HV_UNOP_ABS",
        "__sqrt": "HV_UNOP_SQRT",
        "__ceil": "HV_UNOP_CEIL",
        "__floor": "HV_UNOP_FLOOR",
        "__round": "HV_UNOP_ROUND"
    }

    c_struct = "ControlUnop"
    preamble = "cUnop"

    @classmethod
    def handles_type(clazz, obj_type):
        """Returns true if the object type can be handled by this class
        """
        return obj_type in ControlUnop.__OPERATION_DICT

    @classmethod
    def get_C_header_set(clazz):
        return {"HvControlUnop.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvControlUnop.h", "HvControlUnop.c"}

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [
            "cUnop_onMessage(_c, {1}, m, &cUnop_{0}_sendMessage);".format(
                obj_id,
                ControlUnop.__OPERATION_DICT[obj_type])
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import decimal
import json
import importlib_resources

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class HeavyObject(PdObject):

    heavy_lang_json = importlib_resources.files('hvcc') / 'core/json/heavy.lang.json'
    with open(heavy_lang_json, "r") as f:
        __HEAVY_LANG_OBJS = json.load(f)

    heavy_ir_json = importlib_resources.files('hvcc') / 'core/json/heavy.ir.json'
    with open(heavy_ir_json, "r") as f:
        __HEAVY_IR_OBJS = json.load(f)

    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        PdObject.__init__(self, obj_type, obj_args, pos_x, pos_y)

        # get the object dictionary (note that it is NOT a copy)
        if self.is_hvlang:
            self.__obj_dict = HeavyObject.__HEAVY_LANG_OBJS[obj_type]
        elif self.is_hvir:
            self.__obj_dict = HeavyObject.__HEAVY_IR_OBJS[obj_type]
        else:
            assert False, "{0} is not a Heavy Lang or IR object.".format(obj_type)

        # resolve arguments
        obj_args = obj_args or []
        self.obj_args = {}
        for i, a in enumerate(self.__obj_dict["args"]):
            # if the argument exists (and has been correctly resolved)
            if i < len(obj_args) and obj_args[i] is not None:
                # force the Heavy argument type
                # Catch type errors as early as possible
                try:
                    self.obj_args[a["name"]] = HeavyObject.force_arg_type(
                        obj_args[i],
                        a["value_type"])
                except Exception as e:
                    self.add_error(
                        "Heavy {0} cannot convert argument \"{1}\" with value \"{2}\" to type {3}: {4}".format(
                            obj_type,
                            a["name"],
                            obj_args[i],
                            a["value_type"],
                            str(e)))
            else:
                # the default argument is required
                if a["required"]:
                    self.add_error(
                        "Required argument \"{0}\" to object {1} not present: {2}".format(
                            a["name"],
                            obj_type,
                            obj_args))
                else:
                    # don't worry about supplying a default,
                    # let hv2ir take care of it. pd2hv only passes on the
                    # provided parameters.
                    # self.obj_args[a["name"]] = a["default"]
                    pass

        self.__annotations = {}

        # send/receive, table, etc. must have public scope
        # TODO(mhroth): dirty
        if obj_type in ["table", "__table", "send", "__send", "receive", "__receive"]:
            self.__annotations["scope"] = "public"

    @classmethod
    def force_arg_type(clazz, value, value_type):
        # TODO(mhroth): add support for mixedarray?
        if value_type == "auto":
            try:
                return float(value)
            except Exception:
                return str(value)
        elif value_type == "float":
            return float(value)
        elif value_type == "int":
            return int(decimal.Decimal(value))
        elif value_type == "string":
            return str(value)
        elif value_type == "boolean":
            if isinstance(value, str):
                return value.strip().lower() not in ["false", "f", "0"]
            else:
                return bool(value)
        elif value_type == "floatarray":
            if isinstance(value, list):
                return [float(v) for v in value]
            if isinstance(value, str):
                return [float(v) for v in value.split()]
            else:
                raise Exception(f"Cannot convert value to type floatarray: {value}")
        elif value_type == "intarray":
            if isinstance(value, list):
                return [int(v) for v in value]
            if isinstance(value, str):
                return [int(v) for v in value.split()]
            else:
                raise Exception(f"Cannot convert value to type intarray: {value}")
        elif value_type == "stringarray":
            if isinstance(value, list):
                return [str(v) for v in value]
            if isinstance(value, str):
                return [str(v) for v in value.split()]
            else:
                raise Exception(f"Cannot convert value to type stringarray: {value}")
        else:
            # NOTE(mhroth): if value_type is not a known type or None, that is
            # not necessarily an error. It may simply be that the value should
            # not be resolved to anything other than what it already is.
            # This happens most often with message objects.
            return value

    @property
    def is_hvlang(self):
        return self.obj_type in HeavyObject.__HEAVY_LANG_OBJS

    @property
    def is_hvir(self):
        return self.obj_type in HeavyObject.__HEAVY_IR_OBJS

    def get_inlet_connection_type(self, inlet_index):
        """ Returns the inlet connection type, None if the inlet does not exist.
        """
        # TODO(mhroth): it's stupid that hvlang and hvir json have different data formats here
        if self.is_hvlang:
            if len(self.__obj_dict["inlets"]) > inlet_index:
                return self.__obj_dict["inlets"][inlet_index]["connectionType"]
            else:
                return None
        elif self.is_hvir:
            if len(self.__obj_dict["inlets"]) > inlet_index:
                return self.__obj_dict["inlets"][inlet_index]
            else:
                return None
        else:
            return None

    def get_outlet_connection_type(self, outlet_index):
        """ Returns the outlet connection type, None if the inlet does not exist.
        """
        # TODO(mhroth): it's stupid that hvlang and hvir json have different data formats here
        if self.is_hvlang:
            if len(self.__obj_dict["outlets"]) > outlet_index:
                return self.__obj_dict["outlets"][outlet_index]["connectionType"]
            else:
                return None
        elif self.is_hvir:
            if len(self.__obj_dict["outlets"]) > outlet_index:
                return self.__obj_dict["outlets"][outlet_index]
            else:
                return None
        else:
            return None

    def add_connection(self, c):
        """ Adds a connection, either inlet or outlet, to this object.
        """
        if c.from_id == self.obj_id:
            if self.get_outlet_connection_type(c.outlet_index) is not None:
                self._outlet_connections[str(c.outlet_index)].append(c)
            else:
                self.add_error(
                    f"Connection made from non-existent outlet at {self.obj_type}:{c.outlet_index}.",
                    enum=NotificationEnum.ERROR_UNABLE_TO_CONNECT_OBJECTS)
        elif c.to_id == self.obj_id:
            if self.get_inlet_connection_type(c.inlet_index) is not None:
                self._inlet_connections[str(c.inlet_index)].append(c)
            else:
                self.add_error(
                    f"Connection made to non-existent inlet at [{self.obj_type} {self.obj_args}]:{c.inlet_index}.",
                    enum=NotificationEnum.ERROR_UNABLE_TO_CONNECT_OBJECTS)
        else:
            raise Exception("Adding a connection to the wrong object!")

    def to_hv(self):
        return {
            "type": self.obj_type,
            "args": self.obj_args,
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            },
            "annotations": self.__annotations
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyIrObject import HeavyIrObject


class HIrConvolution(HeavyIrObject):
    """ __conv~f
    """

    def __init__(self, obj_type, args=None, graph=None, annotations=None):
        assert obj_type == "__conv~f"
        HeavyIrObject.__init__(self, obj_type, args=args, graph=graph, annotations=annotations)

    def reduce(self):
        table_obj = self.graph.resolve_object_for_name(
            self.args["table"],
            ["table", "__table"])
        if table_obj is not None:
            self.args["table_id"] = table_obj.id
            return ({self}, [])
        else:
            self.add_error("Cannot find table named \"{0}\" for object {1}.".format(self.args["table"], self))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import jinja2
import os
import shutil
import time
from ..buildjson import buildjson
from ..copyright import copyright_manager


class c2fabric:
    """Generates a DSP component for Fabric.
    """

    @classmethod
    def filter_xcode_copy(clazz, s):
        """Return a copyref hash suitable for use in an Xcode project file.
        """
        return hashlib.md5(s + "_copy").hexdigest().upper()[0:24]

    @classmethod
    def filter_xcode_build(clazz, s):
        """Return a build hash suitable for use in an Xcode project file.
        """
        return hashlib.md5(s + "_build").hexdigest().upper()[0:24]

    @classmethod
    def filter_xcode_fileref(clazz, s):
        """Return a fileref hash suitable for use in an Xcode project file.
        """
        return hashlib.md5(s + "_fileref").hexdigest().upper()[0:24]

    @classmethod
    def filter_templates(clazz, template_name):
        return False if os.path.basename(template_name) in [".DS_Store"] else True

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs,
                patch_name=None, num_input_channels=0, num_output_channels=0,
                copyright=None, verbose=False):

        tick = time.time()

        in_parameter_list = externs["parameters"]["in"]
        out_parameter_list = externs["parameters"]["out"]
        in_event_list = externs["events"]["in"]
        out_event_list = externs["events"]["out"]

        patch_name = patch_name or "heavy"

        copyright = copyright_manager.get_copyright_for_c(copyright)

        # initialise the jinja template environment
        env = jinja2.Environment()
        env.filters["xcode_build"] = c2fabric.filter_xcode_build
        env.filters["xcode_fileref"] = c2fabric.filter_xcode_fileref
        env.filters["xcode_copy"] = c2fabric.filter_xcode_copy
        env.loader = jinja2.FileSystemLoader(
            encoding="utf-8-sig",
            searchpath=[os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")])

        src_out_dir = os.path.join(out_dir, "source")

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over generated C source files
            src_out_dir = os.path.join(out_dir, "source", "heavy")
            shutil.copytree(c_src_dir, src_out_dir)

            files_to_copy = [w.format(patch_name) for w in ["Hv_{0}_FabricDSP.cs", "Hv_{0}_FabricDSPEditor.cs"]]

            # generate files from templates
            for f in env.list_templates(filter_func=c2fabric.filter_templates):
                file_path = os.path.join(out_dir, f)
                file_path = file_path.replace("{{name}}", patch_name)

                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))

                with open(file_path, "w") as g:
                    g.write(env.get_template(f).render(
                        patch_name=patch_name,
                        project_name="Hv_{0}_Fabric".format(patch_name),
                        lib_name=patch_name,
                        num_input_channels=num_input_channels,
                        num_output_channels=num_output_channels,
                        in_parameters=in_parameter_list,
                        out_parameters=out_parameter_list,
                        in_events=in_event_list,
                        out_events=out_event_list,
                        compile_files=os.listdir(src_out_dir),
                        copy_files=files_to_copy,
                        copyright=copyright))

            buildjson.generate_json(
                out_dir,
                android_armv7a_args=["APP_ABI=armeabi-v7a", "-j"],
                linux_x64_args=["-j"],
                macos_x64_args=["-project", "Hv_{0}_Fabric.xcodeproj".format(patch_name),
                                "-arch", "x86_64", "-alltargets"],
                win_x64_args=["/property:Configuration=Release", "/property:Platform=x64",
                              "/t:Rebuild", "Hv_{0}_Fabric.sln".format(patch_name), "/m"],
                win_x86_args=["/property:Configuration=Release", "/property:Platform=x86",
                              "/t:Rebuild", "Hv_{0}_Fabric.sln".format(patch_name), "/m"])

            return {
                "stage": "c2fabric",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2fabric",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from collections import Counter
from collections import OrderedDict
import jinja2
import json
import os
import shutil
import time

from .PrettyfyC import PrettyfyC
from ..copyright import copyright_manager

from .ControlBinop import ControlBinop
from .ControlCast import ControlCast
from .ControlDelay import ControlDelay
from .ControlIf import ControlIf
from .ControlMessage import ControlMessage
from .ControlPack import ControlPack
from .ControlPrint import ControlPrint
from .ControlReceive import ControlReceive
from .ControlRandom import ControlRandom
from .ControlSend import ControlSend
from .ControlSlice import ControlSlice
from .ControlSwitchcase import ControlSwitchcase
from .ControlSystem import ControlSystem
from .ControlTabhead import ControlTabhead
from .ControlTabread import ControlTabread
from .ControlTabwrite import ControlTabwrite
from .ControlUnop import ControlUnop
from .ControlVar import ControlVar
from .HeavyObject import HeavyObject
from .HeavyTable import HeavyTable
from .SignalConvolution import SignalConvolution
from .SignalBiquad import SignalBiquad
from .SignalCPole import SignalCPole
from .SignalDel1 import SignalDel1
from .SignalEnvelope import SignalEnvelope
from .SignalLine import SignalLine
from .SignalLorenz import SignalLorenz
from .SignalMath import SignalMath
from .SignalPhasor import SignalPhasor
from .SignalRPole import SignalRPole
from .SignalSample import SignalSample
from .SignalSamphold import SignalSamphold
from .SignalTabhead import SignalTabhead
from .SignalTabread import SignalTabread
from .SignalTabwrite import SignalTabwrite
from .SignalVar import SignalVar


class ir2c:

    __OBJECT_CLASS_DICT = {
        "__delay": ControlDelay,
        "__if": ControlIf,
        "__print": ControlPrint,
        "__random": ControlRandom,
        "__var": ControlVar,
        "__table": HeavyTable,
        "__cast_b": ControlCast,
        "__cast_f": ControlCast,
        "__cast_s": ControlCast,
        "__message": ControlMessage,
        "__system": ControlSystem,
        "__receive": ControlReceive,
        "__switchcase": ControlSwitchcase,
        "__conv~f": SignalConvolution,
        "__biquad~f": SignalBiquad,
        "__biquad_k~f": SignalBiquad,
        "__env~f": SignalEnvelope,
        "__line~f": SignalLine,
        "__lorenz~f": SignalLorenz,
        "__del1~f": SignalDel1,
        "__tabread~if": SignalTabread,
        "__tabread~f": SignalTabread,
        "__tabreadu~f": SignalTabread,
        "__tabhead~f": SignalTabhead,
        "__tabwrite~f": SignalTabwrite,
        "__tabwrite_stoppable~f": SignalTabwrite,
        "__phasor~f": SignalPhasor,
        "__phasor_k~f": SignalPhasor,
        "__sample~f": SignalSample,
        "__samphold~f": SignalSamphold,
        "__slice": ControlSlice,
        "__send": ControlSend,
        "__tabhead": ControlTabhead,
        "__tabread": ControlTabread,
        "__tabwrite": ControlTabwrite,
        "__pack": ControlPack,
        "__rpole~f": SignalRPole,
        "__cpole~f": SignalCPole
    }

    # the base set of C files necessary for the patch
    __BASE_FILE_SET = {
        "HeavyContextInterface.hpp",
        "HeavyContext.hpp", "HeavyContext.cpp",
        "HvHeavy.h", "HvHeavyInternal.h", "HvHeavy.cpp",
        "HvUtils.h", "HvUtils.c", "HvMath.h",
        "HvMessageQueue.h", "HvMessageQueue.c",
        "HvMessagePool.h", "HvMessagePool.c",
        "HvTable.h", "HvTable.c",
        "HvMessage.h", "HvMessage.c",
        "HvLightPipe.h", "HvLightPipe.c"
    }

    @classmethod
    def filter_hvhash(clazz, x):
        """ Return the hash string of an object.
        """
        return HeavyObject.get_hash_string(x)

    @classmethod
    def filter_extern(clazz, d):
        """ Return a dictionary of objects that are externed.
        """
        return {k: v for k, v in d.items() if v["extern"]}

    @classmethod
    def get_class(clazz, obj_type):
        if SignalMath.handles_type(obj_type):
            return SignalMath
        elif ControlBinop.handles_type(obj_type):
            return ControlBinop
        elif ControlUnop.handles_type(obj_type):
            return ControlUnop
        elif SignalVar.handles_type(obj_type):
            return SignalVar
        elif obj_type in ir2c.__OBJECT_CLASS_DICT:
            return ir2c.__OBJECT_CLASS_DICT[obj_type]
        else:
            raise Exception("No class found for object type \"{0}\".".format(obj_type))

    @classmethod
    def compile(clazz, hv_ir_path, static_dir, output_dir, externs, copyright=None):
        """ Compiles a HeavyIR file into a C.
            Returns a tuple of compile time in seconds, a notification dictionary,
            and a HeavyIR object counter.
        """

        # keep track of the total compile time
        tick = time.time()

        # establish the jinja environment
        env = jinja2.Environment()
        env.filters["hvhash"] = ir2c.filter_hvhash
        env.filters["extern"] = ir2c.filter_extern
        env.loader = jinja2.FileSystemLoader(
            os.path.join(os.path.dirname(__file__), "templates"))

        # read the hv.ir.json file
        with open(hv_ir_path, "r") as f:
            ir = json.load(f)

        # generate the copyright
        copyright = copyright_manager.get_copyright_for_c(copyright)

        #
        # Parse the hv.ir data structure and generate C-language strings.
        #

        # generate set of header files to include
        include_set = set([x for o in ir["objects"].values() for x in ir2c.get_class(o["type"]).get_C_header_set()])

        # generate set of files to add to project
        file_set = set([x for o in ir["objects"].values() for x in ir2c.get_class(o["type"]).get_C_file_set()])
        file_set.update(ir2c.__BASE_FILE_SET)

        # generate object definition and initialisation list
        init_list = []
        free_list = []
        def_list = []
        decl_list = []
        for obj_id in ir["init"]["order"]:
            o = ir["objects"][obj_id]
            obj_class = ir2c.get_class(o["type"])
            init_list.extend(obj_class.get_C_init(o["type"], obj_id, o["args"]))
            def_list.extend(obj_class.get_C_def(o["type"], obj_id))
            free_list.extend(obj_class.get_C_free(o["type"], obj_id, o["args"]))

        impl_list = []
        for x in ir["control"]["sendMessage"]:
            obj_id = x["id"]
            o = ir["objects"][obj_id]
            obj_class = ir2c.get_class(o["type"])
            impl = obj_class.get_C_impl(
                o["type"],
                obj_id,
                x["onMessage"],
                ir2c.get_class,
                ir["objects"])
            impl_list.append("\n".join(PrettyfyC.prettyfy_list(impl)))
            decl_list.extend(obj_class.get_C_decl(o["type"], obj_id, o["args"]))

        # generate static table data initialisers
        table_data_list = []
        for k, v in ir["tables"].items():
            o = ir["objects"][v["id"]]
            obj_class = ir2c.get_class(o["type"])
            table_data_list.extend(obj_class.get_table_data_decl(
                o["type"],
                v["id"],
                o["args"]))

        # generate the list of functions to process
        process_list = []
        for x in ir["signal"]["processOrder"]:
            obj_id = x["id"]
            o = ir["objects"][obj_id]
            process_list.extend(ir2c.get_class(o["type"]).get_C_process(
                o["type"],
                x,
                ir["objects"][obj_id]["args"]))

        #
        # Load the C-language template files and use the parsed strings to fill them in.
        #

        # make the output directory if necessary
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        # the project name to be used as a part of file and function names
        name = ir["name"]["escaped"]

        # ensure that send_receive dictionary is alphabetised by the receiver key
        send_receive = OrderedDict(sorted([(k, v) for k, v in ir["control"]["receivers"].items()], key=lambda x: x[0]))

        # write HeavyContext.h
        with open(os.path.join(output_dir, "Heavy_{0}.hpp".format(name)), "w") as f:
            f.write(env.get_template("Heavy_NAME.hpp").render(
                name=name,
                include_set=include_set,
                decl_list=decl_list,
                def_list=def_list,
                signal=ir["signal"],
                copyright=copyright,
                externs=externs))

        # write C++ implementation
        with open(os.path.join(output_dir, "Heavy_{0}.cpp".format(name)), "w") as f:
            f.write(env.get_template("Heavy_NAME.cpp").render(
                name=name,
                signal=ir["signal"],
                init_list=init_list,
                free_list=free_list,
                impl_list=impl_list,
                send_receive=send_receive,
                send_table=ir["tables"],
                process_list=process_list,
                table_data_list=table_data_list,
                copyright=copyright))

        # write C API, hv_NAME.h
        with open(os.path.join(output_dir, "Heavy_{0}.h".format(name)), "w") as f:
            f.write(env.get_template("Heavy_NAME.h").render(
                name=name,
                copyright=copyright,
                externs=externs))

        # copy static files to output directory
        for f in file_set:
            shutil.copy2(
                src=os.path.join(static_dir, f),
                dst=os.path.join(output_dir, f))

        # generate HeavyIR object counter
        ir_counter = Counter([obj["type"] for obj in ir["objects"].values()])

        return {
            "stage": "ir2c",
            "notifs": {
                "has_error": False,
                "exception": None,
                "errors": []
            },
            "in_dir": os.path.dirname(hv_ir_path),
            "in_file": os.path.basename(hv_ir_path),
            "out_dir": output_dir,
            "out_file": "",
            "compile_time": (time.time() - tick),
            "obj_counter": ir_counter
        }


def main():
    parser = argparse.ArgumentParser(
        description="A Heavy.IR to C-language translator.")
    parser.add_argument(
        "hv_ir_path",
        help="The path to the Heavy.IR file to read.")
    parser.add_argument(
        "--static_dir",
        default="./static",
        help="The path to the static C files.")
    parser.add_argument(
        "--output_dir",
        default="./out",
        help="")
    parser.add_argument(
        "--copyright",
        default=None,
        help="A string indicating the owner of the copyright.")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    results = ir2c.compile(
        args.hv_ir_path,
        args.static_dir,
        args.output_dir,
        args.copyright)

    if args.verbose:
        print("Total ir2c time: {0:.2f}ms".format(results["compile_time"] * 1000))


if __name__ == "__main__":
    main()



# Copyright 2018 Enzien Audio, Ltd. All Rights Reserved.

import os

from .Connection import Connection
from .PdGraph import PdGraph
from .HeavyObject import HeavyObject


class PdLibSignalGraph(PdGraph):

    def __init__(self, obj_args, pd_path, pos_x=0, pos_y=0):
        PdGraph.__init__(self, obj_args, pd_path, pos_x, pos_y)

    def validate_configuration(self):
        """ Auto-detect control only connections to expected signal inlets and
            insert a sig~ (var) object inbetween to ensure it doesn't break.
        """
        # Note(joe): only checking first inlet, should we check all inlets by
        # default? i.e. samphold~ has a 2nd inlet that requires checking
        conns = self._inlet_connections.get("0", [])

        # only do this if no signal connections are present
        if (len([c for c in conns if c.conn_type == "~f>"]) == 0) and len(conns) > 0:

            # add sig~ object to parent graph
            sig_obj = HeavyObject(
                obj_type="var",
                obj_args=[0],
                pos_x=int(self.pos_x),
                pos_y=int(self.pos_y - 5))  # shift upwards a few points
            self.parent_graph.add_object(sig_obj)

            # add connection from sig~ to this abstraction
            c = Connection(sig_obj, 0, self, 0, "~f>")
            self.parent_graph._PdGraph__connections.append(c)  # update the local connections list
            sig_obj.add_connection(c)
            self.add_connection(c)

            # retrieve all control connections
            control_conns = [c for c in conns if c.conn_type == "-->"]

            for old_conn in control_conns:
                # get from obj
                from_obj = old_conn.from_obj

                # add connection from fromobj to new sig
                new_conn = Connection(from_obj, old_conn.outlet_index, sig_obj, 0, "-->")
                self.parent_graph._PdGraph__connections.append(new_conn)
                sig_obj.add_connection(new_conn)
                from_obj.add_connection(new_conn)

                # remove connection from fromobj
                self.parent_graph._PdGraph__connections.remove(old_conn)
                from_obj.remove_connection(old_conn)
                self.remove_connection(old_conn)

        # make sure to call parent validate_configuration()
        PdGraph.validate_configuration(self)

    def __repr__(self):
        return "[{0} {1}]".format(
            os.path.splitext(os.path.basename(self._PdGraph__pd_path))[0],
            " ".join(self.obj_args[1:]))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlSystem(HeavyObject):

    c_struct = "ControlSystem"
    preamble = "cSystem"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvControlSystem.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvControlSystem.h", "HvControlSystem.c"}

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [
            "cSystem_onMessage(_c, NULL, {1}, m, &cSystem_{0}_sendMessage);".format(
                obj_id,
                inlet_index)
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalBiquad(HeavyObject):
    """Handles the biquad~ object.
    """

    c_struct = "SignalBiquad"
    preamble = "sBiquad"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalBiquad.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvSignalBiquad.h", "HvSignalBiquad.c"}

    @classmethod
    def get_C_def(clazz, obj_type, obj_id):
        if obj_type == "__biquad_k~f":
            return ["SignalBiquad_k sBiquad_k_{0};".format(obj_id)]
        elif obj_type == "__biquad~f":
            return ["SignalBiquad sBiquad_s_{0};".format(obj_id)]
        else:
            raise Exception()

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        if obj_type == "__biquad_k~f":
            return ["sBiquad_k_init(&sBiquad_k_{0}, {1}f, {2}f, {3}f, {4}f, {5}f);".format(
                obj_id,
                float(args["ff0"]),
                float(args["ff1"]),
                float(args["ff2"]),
                float(args["fb1"]),
                float(args["fb2"]))]
        elif obj_type == "__biquad~f":
            return ["sBiquad_init(&sBiquad_s_{0});".format(obj_id)]
        else:
            raise Exception()

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return ["sBiquad_k_onMessage(&Context(_c)->sBiquad_k_{0}, {1}, m);".format(
            obj_id,
            inlet_index)]

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, objects):
        if obj_type == "__biquad_k~f":
            return [
                "__hv_biquad_k_f(&sBiquad_k_{0}, VIf({1}), VOf({2}));".format(
                    process_dict["id"],
                    HeavyObject._c_buffer(process_dict["inputBuffers"][0]),
                    HeavyObject._c_buffer(process_dict["outputBuffers"][0])
                )]
        elif obj_type == "__biquad~f":
            return [
                "__hv_biquad_f(&sBiquad_s_{0}, {1}, {2});".format(
                    process_dict["id"],
                    ", ".join(["VIf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["inputBuffers"]]),
                    ", ".join(["VOf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["outputBuffers"]])
                )]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time
import jinja2
from ..copyright import copyright_manager


class c2js:
    """Compiles a directory of C source files into javascript. Requires the
    emscripten library to be installed - https://github.com/kripken/emscripten
    """

    __HV_API = [
        "_hv_{0}_new",
        "_hv_{0}_new_with_options",
        "_hv_delete",
        "_hv_processInline",
        "_hv_getNumInputChannels",
        "_hv_getNumOutputChannels",
        "_hv_samplesToMilliseconds",
        "_hv_setPrintHook",
        "_hv_setSendHook",
        "_hv_sendFloatToReceiver",
        "_hv_sendBangToReceiver",
        "_hv_sendSymbolToReceiver",
        "_hv_stringToHash",
        "_hv_msg_getByteSize",
        "_hv_msg_init",
        "_hv_msg_hasFormat",
        "_hv_msg_setFloat",
        "_hv_msg_getFloat",
        "_hv_msg_getTimestamp",
        "_hv_table_getLength",
        "_hv_table_setLength",
        "_hv_table_getBuffer",
        "_hv_sendMessageToReceiverV"
    ]

    @classmethod
    def run_emscripten(clazz, c_src_dir, out_dir, patch_name, post_js_path):
        """Run the emcc command to compile C source files to a javascript library.
        """
        emcc_path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), "emsdk", "emscripten", "1.37.38", "emcc"))

        emcpp_path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), "emsdk", "emscripten", "1.37.38", "em++"))

        if not os.path.isfile(emcc_path):
            raise Exception("Emscripten emcc command line tool not installed, should be here: " + emcc_path)

        if not os.path.isfile(emcpp_path):
            raise Exception("Emscripten em++ command line tool not installed, should be here: " + emcpp_path)

        c_flags = [
            "-I" + c_src_dir,
            "-DHV_SIMD_NONE",
            "-ffast-math",
            "-DNDEBUG",
            "-Wall"
        ]

        c_src_paths = [os.path.join(c_src_dir, c) for c in os.listdir(c_src_dir) if c.endswith((".c"))]
        cpp_src_paths = [os.path.join(c_src_dir, cpp) for cpp in os.listdir(c_src_dir) if cpp.endswith((".cpp"))]
        obj_paths = []
        cmd = ""

        # compile C files
        for c in c_src_paths:
            obj_path = os.path.splitext(c)[0] + ".o"
            cmd = [emcc_path] + c_flags + ["-c", "-o", obj_path, c]
            subprocess.check_output(cmd)  # run emscripten
            obj_paths += [obj_path]

        # compile C++ files
        for cpp in cpp_src_paths:
            obj_path = os.path.splitext(cpp)[0] + ".o"
            cmd = [emcpp_path] + c_flags + ["-std=c++11"] + ["-c", "-o", obj_path, cpp]
            subprocess.check_output(cmd)  # run emscripten
            obj_paths += [obj_path]

        # exported heavy api methods
        hv_api_defs = ", ".join(["\"{0}\"".format(x) for x in c2js.__HV_API])

        # output path
        asm_js_path = os.path.join(out_dir, "{0}.asm.js".format(patch_name))
        wasm_js_path = os.path.join(out_dir, "{0}.js".format(patch_name))

        linker_flags = [
            "-O3",
            "--memory-init-file", "0",
            "-s", "RESERVED_FUNCTION_POINTERS=2",
            "-s", "EXPORTED_FUNCTIONS=[" + hv_api_defs.format(patch_name) + "]",
            "-s", "MODULARIZE=1",
            "-s", "BINARYEN_TRAP_MODE='clamp'",
            "--post-js", post_js_path
        ]

        # include C/C++ obj files in js library
        cmd = [emcc_path] + obj_paths + linker_flags

        # run emscripten twice!
        subprocess.check_output(  # fallback asm.js build
            cmd + [
                "-s", "EXPORT_NAME='{0}_AsmModule'".format(patch_name),
                "-o", asm_js_path
            ])

        subprocess.check_output(  # WASM
            cmd + [
                "-s", "WASM=1",
                "-s", "EXPORT_NAME='{0}_Module'".format(patch_name),
                "-o", wasm_js_path
            ])

        # clean up
        for o in obj_paths:
            os.remove(o)

        return wasm_js_path

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs,
                patch_name=None, num_input_channels=0, num_output_channels=0,
                copyright=None, verbose=False):

        tick = time.time()

        parameter_list = externs["parameters"]["in"]
        event_list = externs["events"]["in"]

        patch_name = patch_name or "heavy"

        copyright_js = copyright_manager.get_copyright_for_c(copyright)
        copyright_html = copyright_manager.get_copyright_for_xml(copyright)

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        out_dir = os.path.abspath(out_dir)

        try:
            # initialise the jinja template environment
            env = jinja2.Environment()
            env.loader = jinja2.FileSystemLoader(os.path.join(
                os.path.dirname(__file__),
                "template"))

            # generate heavy js wrapper from template
            # Note: this file will be incorporated into the emscripten output
            # and removed afterwards
            post_js_path = os.path.join(out_dir, "hv_wrapper.js")
            with open(post_js_path, "w") as f:
                f.write(env.get_template("hv_wrapper.js").render(
                    name=patch_name,
                    copyright=copyright_js,
                    externs=externs,
                    pool_sizes_kb=externs["memoryPoolSizesKb"]))

            js_path = c2js.run_emscripten(c_src_dir=c_src_dir,
                                          out_dir=out_dir,
                                          patch_name=patch_name,
                                          post_js_path=post_js_path)

            # delete temporary files
            os.remove(post_js_path)

            js_out_file = os.path.basename(js_path)

            # generate index.html from template
            with open(os.path.join(out_dir, "index.html"), "w") as f:
                f.write(env.get_template("index.html").render(
                    name=patch_name,
                    includes=["./{0}".format(js_out_file)],
                    parameters=parameter_list,
                    events=event_list,
                    copyright=copyright_html))

            return {
                "stage": "c2js",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": js_out_file,
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2js",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os

from .PdObject import PdObject
from .HeavyObject import HeavyObject


class HeavyGraph(PdObject):
    def __init__(self, hv_path, obj_args=None, pos_x=0, pos_y=0):
        PdObject.__init__(
            self,
            os.path.basename(hv_path).split(".")[0],
            obj_args,
            pos_x, pos_y)

        # read the heavy graph
        with open(hv_path, "r") as f:
            self.hv_json = json.load(f)

        # parse the heavy data structure to determine the outlet connection type
        outlets = [o for o in self.hv_json["objects"].values() if o["type"] == "outlet"]
        sorted(outlets, key=lambda o: o["args"]["index"])
        self.__outlet_connection_types = [o["args"]["type"] for o in outlets]

        # resolve the arguments
        for i, a in enumerate(self.hv_json["args"]):
            if i < len(self.obj_args):
                arg_value = self.obj_args[i]
            elif a["required"]:
                self.add_error("Required argument \"{0}\" not found.".format(a["name"]))
                continue
            else:
                arg_value = a["default"]

            try:
                arg_value = HeavyObject.force_arg_type(arg_value, a["value_type"])
            except Exception as e:
                self.add_error("Heavy {0} cannot convert argument \"{1}\" with value \"{2}\" to type {3}: {4}".format(
                               self.obj_type,
                               a["name"],
                               arg_value,
                               a["value_type"],
                               str(e)))

            # resolve all arguments for each object in the graph
            for o in self.hv_json["objects"].values():
                for k, v in o["args"].items():
                    # TODO(mhroth): make resolution more robust
                    if v == "$" + a["name"]:
                        o["args"][k] = arg_value

        # reset all arguments, as they have all been resolved
        # any required arguments would break hv2ir as they will no longer
        # be supplied (because they are resolved)
        self.hv_json["args"] = []

    def get_outlet_connection_type(self, outlet_index):
        return self.__outlet_connection_types[outlet_index]

    def to_hv(self):
        return self.hv_json



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from collections import Counter
from collections import defaultdict
import json
import os


class ir2c_perf:

    @classmethod
    def perf(clazz, ir, blocksize=512, mhz=1000, verbose=False):
        # read the hv.ir.json file
        with open(os.path.join(os.path.dirname(__file__), "../../core/json/heavy.ir.json"), "r") as f:
            HEAVY_IR_JSON = json.load(f)

        objects = Counter()
        perf = Counter()
        per_object_perf = defaultdict(Counter)
        for o in ir["signal"]["processOrder"]:
            obj_id = o["id"]
            obj_type = ir["objects"][obj_id]["type"]
            if obj_type in HEAVY_IR_JSON:
                objects[obj_type] += 1
                if "perf" in HEAVY_IR_JSON[obj_type]:
                    c = Counter(HEAVY_IR_JSON[obj_type]["perf"])
                    perf = perf + c
                    per_object_perf[obj_type] = per_object_perf[obj_type] + c
                else:
                    print("{0} requires perf information.".format(obj_type))
            else:
                print("ERROR: Unknown object type {0}".format(obj_type))

        if verbose:
            print("AVX: {0} cycles / {1} cycles per frame".format(perf["avx"], perf["avx"] / 8.0))
            print("     {0} frames @ {1}MHz >= {2:.2f}us".format(
                blocksize,
                mhz,
                blocksize * perf["avx"] / 8.0 / mhz))

            print  # new line

            print("SSE: {0} cycles / {1} cycles per frame".format(perf["sse"], perf["sse"] / 4.0))
            print("     {0} frames @ {1}MHz >= {2:.2f}us".format(
                blocksize,
                mhz,
                blocksize * perf["sse"] / 4.0 / mhz))

            print  # new line

            print("{0:<4} {1:<5} {2:<16} {3}".format("CPU%", "#Objs", "Object Type", "Performance"))
            print("==== ===== ================ ===========")

            # print object in order of highest load
            items = per_object_perf.items()
            # items.sort(key=lambda o: o[1]["avx"], reverse=True)
            for k, v in items:
                print("{2:>2.2g}%  {3:<5} {0:<16} {1}".format(k, v, int(100.0 * v["avx"] / perf["avx"]), objects[k]))

        return per_object_perf


def main():
    parser = argparse.ArgumentParser(
        description="A Heavy.IR to C-language translator.")
    parser.add_argument(
        "hv_ir_path",
        help="The path to the Heavy.IR file to read.")
    parser.add_argument("--mhz", default=1000, type=float, help="the CPU clock frequency in MHz")
    parser.add_argument("--blocksize", default=64, type=int, help="the number of frames per block")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    # read the hv.ir.json file
    with open(args.hv_ir_path, "r") as f:
        ir = json.load(f)

    ir2c_perf.perf(ir, args.blocksize, args.mhz, args.verbose)


if __name__ == "__main__":
    main()



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject

import re


class PdMessageObject(PdObject):

    # only allow dollar argumnets if they are alone
    __RE_DOLLAR = re.compile("\$(\d+)")

    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        assert obj_type == "msg"
        PdObject.__init__(self, "msg", obj_args, pos_x, pos_y)

        self.obj_args = {}

        # parse messages
        # remove prepended slash from $. Heavy does not use that.
        semi_split = obj_args[0].replace("\$", "$").split("\;")
        semi_split = [x for x in semi_split if x]  # remove empty strings

        # parse local messages
        # ensure that empty message are not passed on
        if len(semi_split) > 0:
            self.obj_args["local"] = [li.strip().split() for li in semi_split[0].split("\,") if len(li.strip()) > 0]
        else:
            self.obj_args["local"] = []
            self.add_warning(
                "Message object is empty. Can it be removed?",
                NotificationEnum.WARNING_EMPTY_MESSAGE)

        # heavy does not support messages such as "$1-$2"
        for li in self.obj_args["local"]:
            for m in li:
                x = PdMessageObject.__RE_DOLLAR.search(m)
                if x and len(x.group(0)) < len(m):
                    self.add_error(
                        "Heavy does not yet support message concatenation. "
                        "Dollar arguments must be alone: " + m)

        # parse remote messages
        self.obj_args["remote"] = []
        for li in semi_split[1:]:
            l_split = li.strip().split()
            self.obj_args["remote"].append({
                "receiver": l_split[0],
                "message": l_split[1:]
            })

        if len(self.obj_args["remote"]) > 0:
            self.add_warning(
                "Message boxes don't yet support remote messages. "
                "These messages will be ignored.")

    def to_hv(self):
        return {
            "type": "message",
            "args": self.obj_args,
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from .Connection import Connection
from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdGraph(PdObject):

    def __init__(self, obj_args, pd_path, pos_x=0, pos_y=0):
        assert len(obj_args) > 0, "PdGraph arguments must contain at least dollar zero."
        PdObject.__init__(self, "graph", obj_args, pos_x, pos_y)

        # file location of this graph
        self.__pd_path = pd_path

        self.__objs = []
        self.__connections = []

        self.__inlet_objects = []
        self.__outlet_objects = []

        # the first search path is always the directory of this graph
        self.__declared_paths = [os.path.dirname(pd_path)]

        # heavy graph arguments (added via @hv_arg flag in #X text)
        self.hv_args = []

        # the subpatch name of this graph
        # only used is this graph is actually a subpatch
        self.subpatch_name = None

    @property
    def dollar_zero(self):
        return self.obj_args[0]

    @property
    def is_root(self):
        return self.parent_graph is None

    @property
    def is_subpatch(self):
        return self.parent_graph.__pd_path == self.__pd_path if not self.is_root else False

    def add_object(self, obj):
        obj.parent_graph = self
        self.__objs.append(obj)

        if obj.obj_type in ["inlet", "inlet~"]:
            self.__inlet_objects.append(obj)
            # set correct let index by sorting on x position
            self.__inlet_objects.sort(key=lambda o: o.pos_x)
            for i, o in enumerate(self.__inlet_objects):
                o.let_index = i
        elif obj.obj_type in ["outlet", "outlet~"]:
            self.__outlet_objects.append(obj)
            self.__outlet_objects.sort(key=lambda o: o.pos_x)
            for i, o in enumerate(self.__outlet_objects):
                o.let_index = i

    def add_parsed_connection(self, from_index, from_outlet, to_index, to_inlet):
        """ Add a connection to the graph which has been parsed externally.
        """
        try:
            # when connecting a Heavy object that allows mixed connection types,
            # try to immediately resolve to a non-mixed type based on the
            # object that it is connected to.
            connection_type = self.__objs[from_index].get_outlet_connection_type(from_outlet)
            if connection_type == "-~>":
                connection_type = self.__objs[to_index].get_inlet_connection_type(to_inlet)

            # make the connection
            c = Connection(
                self.__objs[from_index], from_outlet,
                self.__objs[to_index], to_inlet,
                connection_type)
            self.__connections.append(c)  # update the local connections list

            # allow the connected objects to keep track of their own connections
            # (generally used for reporting and validation purposes)
            self.__objs[from_index].add_connection(c)
            self.__objs[to_index].add_connection(c)
        except Exception:
            self.add_error("There was an error while connecting two objects. "
                           "Have all objects been correctly instantiated? "
                           "Have all inlets and outlets been declared?",
                           NotificationEnum.ERROR_UNABLE_TO_CONNECT_OBJECTS)

    def add_hv_arg(self, arg_index, name, value_type, default_value, required):
        """ Add a Heavy argument to the graph. Indicies are from zero (not one, like Pd).
        """
        # ensure that self.hv_args is big enough, as heavy arguments are not
        # necessarily added in the natural order
        while arg_index >= len(self.hv_args):
            self.hv_args.append(None)

        self.hv_args[arg_index] = {
            "name": name,
            "description": "",
            "value_type": value_type,
            "default": default_value,
            "required": required
        }

    def get_inlet_connection_type(self, inlet_index):
        return self.__inlet_objects[inlet_index].get_inlet_connection_type()

    def get_outlet_connection_type(self, outlet_index):
        return self.__outlet_objects[outlet_index].get_outlet_connection_type()

    def validate_configuration(self):
        if self.is_root:
            if any((o.obj_type in ["inlet~", "outlet~"]) for o in self.__objs):
                self.add_error(
                    "Top-level graphs may not contain inlet~ or outlet~ objects. "
                    "Use adc~ and dac~.",
                    NotificationEnum.ERROR_NO_TOPLEVEL_SIGNAL_LETS)
            if any((o.obj_type in ["inlet", "outlet"]) for o in self.__objs):
                self.add_warning(
                    "Control inlets and outlets in top-level graphs don't do "
                    "anything. Use receive and send objects.")

        # validate all object recursively
        for o in self.__objs:
            o.validate_configuration()

    def is_abstraction_on_call_stack(self, abs_path):
        """ Returns True if the given abstraction name is already on the call
            stack (i.e. it is currently being parsed). This function is used to
            detect recursion within abstractions.
        """
        if self.__pd_path == abs_path:
            return True
        elif not self.is_root:
            return self.parent_graph.is_abstraction_on_call_stack(abs_path)
        else:
            return False

    def get_notices(self):
        notices = PdObject.get_notices(self)
        for o in self.__objs:
            n = o.get_notices()
            notices["warnings"].extend(n["warnings"])
            notices["errors"].extend(n["errors"])

        # remove ERROR_EXCEPTION if there are already other errors.
        # The exception is always a result of some other error
        if any((n["enum"] != NotificationEnum.ERROR_EXCEPTION) for n in notices["errors"]):
            notices["errors"] = [n for n in notices["errors"] if n["enum"] != NotificationEnum.ERROR_EXCEPTION]

        return notices

    def get_graph_heirarchy(self):
        """ Returns the "path" of this graph, indicating where it is in the
            graph heirarchy (i.e. with file names, etc.)
        """
        return [str(self)] if self.is_root else \
            self.parent_graph.get_graph_heirarchy() + [str(self)]

    def get_depth(self):
        """ Returns the depth of this graph, with the root being at 1.
        """
        return 1 if self.is_root else (1 + self.parent_graph.get_depth())

    def to_hv(self, export_args=False):
        # NOTE(mhroth): hv_args are not returned. Because all arguments have
        # been resolved, no arguments are otherwise passed. hv2ir would break
        # on required arguments that are not passed to the graph
        assert all(a is not None for a in self.hv_args), "Graph is missing a @hv_arg."
        return {
            "type": "graph",
            "imports": [],
            "args": self.hv_args if export_args else [],
            "objects": {o.obj_id: o.to_hv() for o in self.__objs},
            "connections": [c.to_hv() for c in self.__connections],
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }

    def __repr__(self):
        return self.subpatch_name or os.path.basename(self.__pd_path)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import Counter
from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdRouteObject(PdObject):
    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        assert obj_type == "route"
        PdObject.__init__(self, obj_type, obj_args, pos_x, pos_y)
        if len(obj_args) == 0:
            self.add_error("At least one argument is required.")
        # NOTE(joe): disabling this warning as it can be quite annoying.
        # if len(set(obj_args) & set(["bang", "list", "float", "symbol"])) > 0:
        #     self.add_warning(
        #         "Heavy interprets route arguments such as \"bang\", \"list\", \"float\", "
        #         "and \"symbol\" literally. They cannot be used to filter generic "
        #         "messages as in Pd.")
        if len(set(obj_args)) != len(obj_args):
            c = Counter(obj_args).most_common(1)
            self.add_error(
                "All arguments to [route] must be unique. Argument \"{0}\" is "
                "repeated {1} times.".format(c[0][0], c[0][1]),
                NotificationEnum.ERROR_UNIQUE_ARGUMENTS_REQUIRED)

        # convert to obj_args to mixedarray, such that correct switchcase hash
        # is generated
        for i, a in enumerate(self.obj_args):
            try:
                self.obj_args[i] = float(a)
            except Exception:
                pass

    def validate_configuration(self):
        if len(self._inlet_connections.get("1", [])) > 0:
            self.add_warning(
                "The right inlet of route is not supported. "
                "It will not do anything.")

    def to_hv(self):
        """Creates a graph dynamically based on the number of arguments.
        An unconnected right inlet is added.

        [inlet]                                         [inlet]
        |
        [@hv_obj switchcase [arg list (N elements)]           ]
        |                             |                       |
        [@hv_obj __slice 1 -1]        [@hv_obj __slice 1 -1]  |
        |        |                    |          |            |
        [outlet_0]                    [outlet_N-1]            [outlet_right]
        """

        route_graph = {
            "type": "graph",
            "imports": [],
            "args": [],
            "objects": {
                "inlet": {
                    "type": "inlet",
                    "args": {
                        "type": "-->",
                        "index": 0
                    },
                    "properties": {"x": 0, "y": 0}
                },
                "inlet_right": {
                    "type": "inlet",
                    "args": {
                        "type": "-->",
                        "index": 1
                    },
                    "properties": {"x": 0, "y": 0}
                },
                "switchcase": {
                    "type": "__switchcase",
                    "args": {
                        "cases": self.obj_args
                    },
                    "properties": {"x": 0, "y": 0}
                },
                "outlet_right": {
                    "type": "outlet",
                    "args": {
                        "type": "-->",
                        "index": len(self.obj_args)
                    },
                    "properties": {"x": 0, "y": 0}
                },
            },
            "connections": [
                {
                    "from": {"id": "inlet", "outlet": 0},
                    "to": {"id": "switchcase", "inlet": 0},
                    "type": "-->"
                },
                {
                    "from": {"id": "switchcase", "outlet": len(self.obj_args)},
                    "to": {"id": "outlet_right", "inlet": 0},
                    "type": "-->"
                }
            ],
            "properties": {"x": self.pos_x, "y": self.pos_y}
        }

        # add slices to graph
        for i, a in enumerate(self.obj_args):
            # add slices to graph
            route_graph["objects"]["slice_{0}".format(i)] = {
                "type": "slice",
                "args": {
                    "index": 1,
                    "length": -1
                },
                "properties": {"x": 0, "y": 0}
            }

            # add outlets to graph
            route_graph["objects"]["outlet_{0}".format(i)] = {
                "type": "outlet",
                "args": {
                    "type": "-->",
                    "index": i
                },
                "properties": {"x": 0, "y": 0}
            }

            # add connection from switchcase to slice
            route_graph["connections"].append({
                "from": {"id": "switchcase", "outlet": i},
                "to": {"id": "slice_{0}".format(i), "inlet": 0},
                "type": "-->"
            })

            # add connection from slice outlets 0 and 1 to outlet
            route_graph["connections"].append({
                "from": {"id": "slice_{0}".format(i), "outlet": 0},
                "to": {"id": "outlet_{0}".format(i), "inlet": 0},
                "type": "-->"
            })
            route_graph["connections"].append({
                "from": {"id": "slice_{0}".format(i), "outlet": 1},
                "to": {"id": "outlet_{0}".format(i), "inlet": 0},
                "type": "-->"
            })

        return route_graph



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os
import shutil
import time
import jinja2
from ..copyright import copyright_manager


class c2pdext:
    """Generates a Pure Data external wrapper for a given patch.
    """

    @classmethod
    def filter_max(clazz, i, j):
        """Calculate the maximum of two integers.
        """
        return max(int(i), int(j))

    @classmethod
    def filter_xcode_build(clazz, s):
        """Return a build hash suitable for use in an Xcode project file.
        """
        s = f"{s}_build"
        s = hashlib.md5(s.encode('utf-8'))
        s = s.hexdigest().upper()[0:24]
        return s

    @classmethod
    def filter_xcode_fileref(clazz, s):
        """Return a fileref hash suitable for use in an Xcode project file.
        """
        s = f"{s}_fileref"
        s = hashlib.md5(s.encode('utf-8'))
        s = s.hexdigest().upper()[0:24]
        return s

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs,
                patch_name=None, ext_name=None,
                num_input_channels=0, num_output_channels=0,
                copyright=None, verbose=False):

        tick = time.time()
        receiver_list = externs["parameters"]["in"]

        copyright = copyright_manager.get_copyright_for_c(copyright)

        patch_name = patch_name or "heavy"
        ext_name = ext_name or (patch_name + "~")
        struct_name = patch_name + "_tilde"

        # ensure that the output directory does not exist
        out_dir = os.path.abspath(out_dir)
        if os.path.exists(out_dir):
            shutil.rmtree(out_dir)

        # copy over generated C source files
        shutil.copytree(c_src_dir, out_dir)

        # copy over static files
        shutil.copy(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "static", "m_pd.h"),
            out_dir)

        try:
            # initialise the jinja template environment
            env = jinja2.Environment()
            env.filters["max"] = c2pdext.filter_max
            env.filters["xcode_build"] = c2pdext.filter_xcode_build
            env.filters["xcode_fileref"] = c2pdext.filter_xcode_fileref
            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # generate Pd external wrapper from template
            pdext_path = os.path.join(out_dir, "{0}.c".format(struct_name))
            with open(pdext_path, "w") as f:
                f.write(env.get_template("pd_external.c").render(
                    name=patch_name,
                    struct_name=struct_name,
                    display_name=ext_name,
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    copyright=copyright))

            # generate Xcode project
            xcode_path = os.path.join(out_dir, "{0}.xcodeproj".format(struct_name))
            os.mkdir(xcode_path)  # create the xcode project bundle
            pbxproj_path = os.path.join(xcode_path, "project.pbxproj")

            # generate list of source files
            files = [g for g in os.listdir(out_dir) if g.endswith((".h", ".hpp", ".c", ".cpp"))]

            # render the pbxproj file
            with open(pbxproj_path, "w") as f:
                f.write(env.get_template("project.pbxproj").render(
                    name=ext_name,
                    files=files))

            return {
                "stage": "c2pdext",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": os.path.basename(pdext_path),
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2pdext",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": os.path.basename(pdext_path),
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdTriggerObject(PdObject):
    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        assert obj_type in ["trigger", "t"]
        PdObject.__init__(self, "trigger", obj_args, pos_x, pos_y)

        # convert all numeric casts to "f"
        for i, a in enumerate(self.obj_args):
            self.obj_args[i] = "f" if PdTriggerObject.__is_float(a) else a

        if len(self.obj_args) == 0:
            self.obj_args = ["b", "b"]
            self.add_warning("A trigger with no arguments defualts to [t b b].")
        if not (set(obj_args) <= set(["a", "f", "s", "b"])):
            self.add_error(
                "Heavy only supports arguments 'a', 'f', 's', and 'b'.",
                NotificationEnum.ERROR_TRIGGER_ABFS)

    def to_hv(self):
        return {
            "type": "sequence",
            "args": {
                "casts": self.obj_args
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }

    @classmethod
    def __is_float(clazz, x):
        """ Returns True if the input can be converted to a float. False otherwise.
        """
        try:
            float(x)
            return True
        except Exception:
            return False



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os

from hvcc.core.hv2ir.Connection import Connection
from hvcc.core.hv2ir.HeavyException import HeavyException
from hvcc.core.hv2ir.HeavyLangObject import HeavyLangObject


class HeavyIrObject(HeavyLangObject):
    """ Intermediate Representation (IR) objects are atomic and have
        strictly defined interfaces and types. These are generally defined in
        the file heavy.ir.json.
    """

    # load the HeavyIR object definitions
    with open(os.path.join(os.path.dirname(__file__), "../json/heavy.ir.json"), "r") as f:
        __HEAVY_OBJS_IR_DICT = json.load(f)

    def __init__(self, obj_type, args=None, graph=None, num_inlets=-1, num_outlets=-1, annotations=None):
        # allow the number of inlets and outlets to be overridden
        num_inlets = len(HeavyIrObject.__HEAVY_OBJS_IR_DICT[obj_type]["inlets"]) if num_inlets < 0 else num_inlets

        num_outlets = len(HeavyIrObject.__HEAVY_OBJS_IR_DICT[obj_type]["outlets"]) if num_outlets < 0 else num_outlets

        HeavyLangObject.__init__(self, obj_type, args, graph, num_inlets, num_outlets, annotations)

        # resolve arguments and fill in missing defaults for HeavyIR objects
        self.__resolve_default_ir_args()

        # the list of signal buffers at the inlets and outlets
        # these are filled in by HeavyGraph.assign_signal_buffers()
        self.inlet_buffers = [("zero", 0)] * self.num_inlets
        self.outlet_buffers = [("zero", 0)] * self.num_outlets

        # True if this object has already been ordered in the signal chain
        self.__is_ordered = False

    def __resolve_default_ir_args(self):
        """ Resolves missing default arguments. Also checks to make sure that all
            required arguments are present.
        """
        if self.type in HeavyIrObject.__HEAVY_OBJS_IR_DICT:
            for arg in self.__obj_desc.get("args", []):
                if arg["name"] not in self.args:
                    # if a defined argument is not in the argument dictionary
                    if not arg["required"]:
                        # if the argument is not required, use the default
                        self.args[arg["name"]] = arg["default"]
                    else:
                        self.add_error("Required argument \"{0}\" not present for object {1}.".format(
                            arg["name"],
                            self))
                else:
                    # enforce argument types.
                    # if the default argument is null, don't worry about about the arg
                    if arg["default"] is not None:
                        self.args[arg["name"]] = HeavyLangObject.force_arg_type(
                            self.args[arg["name"]],
                            arg["value_type"],
                            self.graph)

    @classmethod
    def is_ir(clazz, obj_type):
        """Returns true if the type is an IR object. False otherwise.
        """
        return obj_type in HeavyIrObject.__HEAVY_OBJS_IR_DICT

    @property
    def does_process_signal(self):
        """Returns True if this object processes a signal. False otherwise.
        """
        return self.__obj_desc["ir"]["signal"]

    @property
    def __obj_desc(self):
        """ Returns the original HeavyIR object description.
        """
        return HeavyIrObject.__HEAVY_OBJS_IR_DICT[self.type]

    def inlet_requires_signal(self, inlet_index=0):
        """ Returns True if the indexed inlet requires a signal connection. False otherwise.
        """
        return self.__obj_desc["inlets"][inlet_index] in ["~i>", "~f>"]

    def outlet_requires_signal(self, inlet_index=0):
        """ Returns True if the indexed outlet requires a signal connection. False otherwise.
        """
        return self.__obj_desc["outlets"][inlet_index] in ["~i>", "~f>"]

    def reduce(self):
        # A Heavy IR object is already reduced. Returns itself and no connection changes.
        return ({self}, [])

    def get_parent_order(self):
        """ Returns a list of all objects in process order, with this object at the end.
        """
        if self.__is_ordered:
            return []
        else:
            self.__is_ordered = True
            if self.is_root():
                return [self]
            else:
                order_list = []
                for c in [c for inlet in self.inlet_connections for c in inlet]:
                    order_list.extend(c.from_object.get_parent_order())
                order_list.append(self)
                return order_list

    def assign_signal_buffers(self, buffer_pool):
        # assign the inlet buffers
        for cc in self.inlet_connections:
            cc = [c for c in cc if c.is_signal]  # only need to deal with signal connections
            if len(cc) == 0:
                continue
            if len(cc) == 1:
                c = cc[0]  # get the connection

                # get the buffer at the outlet of the connected object
                buf = c.from_object.outlet_buffers[c.outlet_index]

                # assign the buffer to the inlet of this object
                self.inlet_buffers[c.inlet_index] = buf

                # decrease the retain count of the buffer
                buffer_pool.release_buffer(buf)
            else:
                raise HeavyException("This object has {0} (> 1) signal inputs.".format(len(cc)))

        # assign the output buffers
        exclude_set = set()

        for i in range(self.num_outlets):
            # buffers are assigned even if the outlet has no connections.
            # The buffer will still be filled. However, if the buffer has already
            # been set (i.e. non-zero) (e.g. in the case of dac~),
            # then we skip this set

            connection_type = self._resolved_outlet_type(outlet_index=i)
            if Connection.is_signal_type(connection_type) and self.outlet_buffers[i][0] == "zero":
                b = buffer_pool.get_buffer(
                    connection_type,
                    len(self.outlet_connections[i]),
                    exclude_set)
                self.outlet_buffers[i] = b

                # if the buffer has no dependencies, make sure that it isn't reused
                # right away. All outlets should have independent buffers.
                if len(self.outlet_connections[i]) == 0:
                    exclude_set.add(b)

    def _resolved_outlet_type(self, outlet_index=0):
        """ Returns the connection type at the given outlet.
            This information is always well-defined for IR objects.
        """
        return self.__obj_desc["outlets"][outlet_index]

    #
    # Intermediate Representation generators
    #

    def get_object_dict(self):
        """ Returns a dictionary of all constituent low-level objects,
            indexed by id, including their arguments and type.
        """
        return {
            self.id: {
                "args": self.args,
                "type": self.type
            }
        }

    def get_ir_init_list(self):
        """ Returns a list of all object id for obejcts that need initialisation.
        """
        return [self.id] if self.__obj_desc["ir"]["init"] else []

    def get_ir_on_message(self, inlet_index=0):
        """ Returns an array of dictionaries containing the information for the
            corresponding on_message call.
        """
        return [{
            "id": self.id,
            "inletIndex": inlet_index
        }]

    def get_ir_control_list(self):
        """ Returns the intermediate representation for object control functions.
            Basically, does sendMessage() need to be written?
        """
        if self.__obj_desc["ir"]["control"]:
            on_message_list = []
            for connections in self.outlet_connections:
                on_messages_let = []
                # only look at control connections
                for c in [c for c in connections if c.is_control]:
                    on_messages_let.extend(c.to_object.get_ir_on_message(c.inlet_index))
                on_message_list.append(on_messages_let)
            return [{
                "id": self.id,
                "onMessage": on_message_list
            }]
        else:
            return []

    def get_ir_signal_list(self):
        """ Returns the intermediate representation for object process functions.
            Only outputs buffer information for lets that require a signal.
        """
        # we assume that this method will only be called on signal objects
        assert self.__obj_desc["ir"]["signal"]

        return [{
            "id": self.id,
            "inputBuffers": [
                {"type": b[0], "index": b[1]} for i, b in enumerate(self.inlet_buffers)
                if self.inlet_requires_signal(i)],
            "outputBuffers": [
                {"type": b[0], "index": b[1]} for i, b in enumerate(self.outlet_buffers)
                if self.outlet_requires_signal(i)]
        }]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os
import shutil
import time
import jinja2
from ..buildjson import buildjson
from ..copyright import copyright_manager


class c2wwise:
    """Generates a plugin wrapper for Audiokinetic's Wwise game audio middleware
    platform.
    """

    @classmethod
    def filter_xcode_build(clazz, s):
        """Return a build hash suitable for use in an Xcode project file.
        """
        return hashlib.md5(s + "_build").hexdigest().upper()[0:24]

    @classmethod
    def filter_xcode_fileref(clazz, s):
        """Return a fileref hash suitable for use in an Xcode project file.
        """
        return hashlib.md5(s + "_fileref").hexdigest().upper()[0:24]

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs,
                patch_name=None, num_input_channels=0, num_output_channels=0,
                copyright=None, verbose=False):

        tick = time.time()

        in_parameter_list = externs["parameters"]["in"]
        out_parameter_list = externs["parameters"]["out"]
        event_list = externs["events"]["in"]
        table_list = externs["tables"]

        patch_name = patch_name or "heavy"

        copyright_c = copyright_manager.get_copyright_for_c(copyright)
        copyright_xml = copyright_manager.get_copyright_for_xml(copyright)

        wwise_sdk_version = "2017.2.2.6553"

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        templates_dir = os.path.join(os.path.dirname(__file__), "templates")
        plugin_type = "Source" if num_input_channels == 0 else "FX"
        plugin_id = int(hashlib.md5(patch_name).hexdigest()[:4], 16) & 0x7FFF  # unique id from patch name [0...32767]

        env = jinja2.Environment()
        env.filters["xcode_build"] = c2wwise.filter_xcode_build
        env.filters["xcode_fileref"] = c2wwise.filter_xcode_fileref
        env.loader = jinja2.FileSystemLoader(
            encoding="utf-8-sig",
            searchpath=[templates_dir])

        try:
            if plugin_type == "FX":
                if num_input_channels > 2:
                    raise Exception("Wwise FX plugins support a maximum of 2 channels (i.e. [adc~ 1] or [adc~ 1 2]).")
                if num_input_channels != num_output_channels:
                    raise Exception("Wwise FX plugins require the same input/output channel"
                                    "configuration (i.e. [adc~ 1] -> [dac~ 1]).")

            # copy over generated C source files
            patch_src_dir = os.path.join(out_dir, "source", "heavy")
            if os.path.exists(patch_src_dir):
                shutil.rmtree(patch_src_dir)
            shutil.copytree(c_src_dir, patch_src_dir)

            # template all source files
            src_extns = ["h", "hpp", "c", "cpp", "xml", "def", "rc", "plist"]
            for f in env.list_templates(extensions=src_extns):
                file_dir = os.path.join(out_dir, os.path.dirname(f))
                file_name = os.path.basename(f)

                # ignore source plugin files if type is FX
                if (file_dir.endswith("plugin_source") and plugin_type != "Source"):
                    continue

                # ignore fx plugin files if type is Source
                elif (file_dir.endswith("plugin_fx") and plugin_type != "FX"):
                    continue

                # static files
                if file_name in ["stdafx.h", "Info.plist"]:
                    file_path = os.path.join(file_dir, file_name)

                    if not os.path.exists(os.path.dirname(file_path)):
                        os.makedirs(os.path.dirname(file_path))

                    shutil.copyfile(os.path.join(templates_dir, f), file_path)

                # templated files
                else:
                    file_name = file_name.replace("{{name}}", patch_name)
                    file_name = file_name.replace("{{type}}", plugin_type)
                    file_path = os.path.join(file_dir, file_name)

                    if not os.path.exists(os.path.dirname(file_path)):
                        os.makedirs(os.path.dirname(file_path))

                    with open(file_path, "w") as g:
                        g.write(env.get_template(f).render(
                            name=patch_name,
                            parameters=in_parameter_list,
                            sends=out_parameter_list,
                            events=event_list,
                            tables=table_list,
                            pool_sizes_kb=externs["memoryPoolSizesKb"],
                            plugin_type=plugin_type,
                            plugin_id=plugin_id,
                            copyright=copyright_xml if file_name.endswith(".xml") else copyright_c))

            files = [f for f in os.listdir(patch_src_dir)]

            # template all ide project files
            for f in env.list_templates(extensions=["vcxproj", "sln", "pbxproj", "xcconfig"]):

                # xcode projects are structured differently to visual studio projects
                if f.endswith("pbxproj"):
                    xcode_proj_dir = os.path.splitext(os.path.dirname(f))
                    file_path = os.path.join(
                        out_dir,
                        os.path.dirname(xcode_proj_dir[0]),
                        f"Hv_ {patch_name}_Wwise {plugin_type}"
                        f" {os.path.basename(xcode_proj_dir[0])} {xcode_proj_dir[1]}",
                        os.path.basename(f))
                else:
                    file_path = f.replace("{{name}}", patch_name)
                    file_path = file_path.replace("{{type}}", plugin_type)
                    file_path = os.path.join(out_dir, file_path.replace("{{type}}", plugin_type))

                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))

                with open(file_path, "w") as g:
                    g.write(env.get_template(f).render(
                        name=patch_name,
                        parameters=in_parameter_list,
                        sends=out_parameter_list,
                        events=event_list,
                        tables=table_list,
                        plugin_type=plugin_type,
                        plugin_id=plugin_id,
                        wwise_version=wwise_sdk_version,
                        msbuild_version="140",
                        files=files))

            # linux makefile
            linux_makefile = os.path.join(out_dir, "linux", "Makefile")

            if not os.path.exists(os.path.dirname(linux_makefile)):
                os.makedirs(os.path.dirname(linux_makefile))

            with open(linux_makefile, "w") as g:
                g.write(env.get_template("linux/Makefile").render(
                    name=patch_name,
                    plugin_type=plugin_type,
                    plugin_id=plugin_id,
                    files=files,
                    wwise_version=wwise_sdk_version))

            proj_name = "Hv_{0}_Wwise{1}Plugin".format(patch_name, plugin_type)

            buildjson.generate_json(
                out_dir,
                ios_armv7a_args=[
                    "-arch", "armv7s",
                    "-target", "{0}_iOS".format(proj_name),
                    "-project", "{0}.xcodeproj".format(proj_name)],
                linux_x64_args=["-j"],
                macos_x64_args=[
                    "-arch", "x86_64",
                    "-target", "{0}".format(proj_name),
                    "-project", "{0}.xcodeproj".format(proj_name)],
                win_x64_args=[
                    "/property:Configuration=Release",
                    "/property:Platform=x64",
                    "/t:Rebuild", "/m",
                    "{0}.sln".format(proj_name)],
                win_x86_args=[
                    "/property:Configuration=Release",
                    "/property:Platform=x86",
                    "/t:Rebuild", "/m",
                    "{0}.sln".format(proj_name)])

            return {
                "stage": "c2wwise",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2wwise",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

# http://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python


class NotificationEnum(Enum):
    """ These enumerations refer to all possible warnings and errors produced by pd2hv.
    """

    # Warnings

    # an unspecified generic warning
    WARNING_GENERIC = 1000

    # an empty object has been added to a Pd graph
    WARNING_EMPTY_OBJECT = 1001

    # heavy only supports the -path flag to the Pd declare object
    WARNING_DECLARE_PATH = 1002

    # the object does nothing (and will likely be removed)
    WARNING_USELESS_OBJECT = 1003

    # the message box is empty and does nothing
    WARNING_EMPTY_MESSAGE = 1004

    # Errors

    # an unspecified generic error
    ERROR_GENERIC = 2000

    # the parser doesn't know how to handle the object
    ERROR_UNKNOWN_OBJECT = 2001

    # top-level graphs may not contain inlet~ or outlet~ objects
    ERROR_NO_TOPLEVEL_SIGNAL_LETS = 2002

    # only arguments 'a', 'f', 's', and 'b' are supported for trigger object
    ERROR_TRIGGER_ABFS = 2003

    # only arguments 'f', 'float' and numeric values are supported for pack object
    ERROR_PACK_FLOAT_ARGUMENTS = 2003

    # a value cannot be resolved because the required argument is missing
    ERROR_MISSING_REQUIRED_ARGUMENT = 2004

    # unique arguments are required for this object
    ERROR_UNIQUE_ARGUMENTS_REQUIRED = 2005

    # two objects cannot be connected
    ERROR_UNABLE_TO_CONNECT_OBJECTS = 2006

    # a particular connection (type?) is not supported
    ERROR_UNSUPPORTED_CONNECTION = 2007

    # Exception

    # the exception that was raised
    ERROR_EXCEPTION = 3000



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

import argparse
import json
import os
import time

from .MaxParser import MaxParser


class max2hv:

    @classmethod
    def compile(clazz, max_path, hv_dir, search_paths=None, verbose=False):
        tick = time.time()

        max_graph = MaxParser.graph_from_file(max_path)

        if not os.path.exists(hv_dir):
            os.makedirs(hv_dir)

        hv_file = os.path.basename(max_path).split(".")[0] + ".hv.json"
        hv_path = os.path.join(hv_dir, hv_file)
        with open(hv_path, "w") as f:
            if verbose:
                f.write(json.dumps(
                    max_graph.to_hv(),
                    sort_keys=True,
                    indent=2,
                    separators=(",", ": ")))
            else:
                f.write(json.dumps(max_graph.to_hv()))

        return {
            "stage": "max2hv",
            "notifs": {
                "has_error": False,
                "exception": None,
                "errors": []
            },
            "in_dir": os.path.dirname(max_path),
            "in_file": os.path.basename(max_path),
            "out_dir": hv_dir,
            "out_file": hv_file,
            "compile_time": (time.time() - tick)
        }


def main():
    parser = argparse.ArgumentParser(
        description="Converts a Max patch into the Heavy language format.")
    parser.add_argument(
        "max_path",
        help="The Max patch to convert to Heavy")
    parser.add_argument(
        "hv_dir",
        help="Directory to store generated Heavy patches.")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show debugging information",
        action="count")
    args = parser.parse_args()

    args.max_path = os.path.abspath(os.path.expanduser(args.max_path))
    args.hv_dir = os.path.abspath(os.path.expanduser(args.hv_dir))

    result = max2hv.compile(
        max_path=args.max_path,
        hv_dir=args.hv_dir,
        search_paths=None,
        verbose=args.verbose)

    if result["notifs"].get("has_error", False):
        print(result["notifs"]["errors"])

    if args.verbose:
        print("Heavy file written to", os.path.join(result["output_dir"], result["output_file"]))
        print("Total max2hv compile time: {0:.2f}ms".format(result["compile_time"] * 1000))


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyLangObject import HeavyLangObject
from .HIrReceive import HIrReceive


class HLangReceive(HeavyLangObject):
    """ Translates HeavyLang object [receive] to HeavyIR [__receive if control,
        otherwise for signal connections it will remove the send~/receive~ objects
        and reorder the graph.
    """

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(self, "receive", args, graph, annotations=annotations)

    def reduce(self):
        if self.has_outlet_connection_format(["c"]):
            x = HIrReceive("__receive", self.args, annotations=self.annotations)
            return ({x}, self.get_connection_move_list(x))

        elif self.has_outlet_connection_format(["f"]):
            # this case should be handled already in HeavyGraph.remap_send_receive()
            # clean up just in case
            return (set(), [])

        elif self.has_outlet_connection_format("_"):
            self.add_warning("receive~ object doesn't have any outputs, this one is being removed.")
            return (set(), [])

        else:
            fmt = self._get_connection_format(self.outlet_connections)
            self.add_error("Unknown outlet configuration: {0}".format(fmt))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject


class HLangVar(HeavyLangObject):

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(self, "var", args, graph, annotations=annotations)

    def reduce(self):
        if self.has_inlet_connection_format("_") and self.has_outlet_connection_format("f"):
            x = HeavyIrObject("__var_k~f", self.args)
        elif self.has_inlet_connection_format("_") and self.has_outlet_connection_format("i"):
            x = HeavyIrObject("__var_k~i", self.args)
        elif self.has_inlet_connection_format(["_", "c"]) and self.has_outlet_connection_format("f"):
            x = HeavyIrObject("__var~f", self.args)
            x.id = self.id
            y = HeavyIrObject("__varread~f", {"var_id": self.id})

            move_list = []
            for c in [c for cc in self.inlet_connections for c in cc]:
                move_list.append((c, [c.copy(to_object=x)]))
            for c in [c for cc in self.outlet_connections for c in cc]:
                move_list.append((c, [c.copy(from_object=y)]))

            return ({x, y}, move_list)
        elif self.has_inlet_connection_format(["_", "c"]) and self.has_outlet_connection_format("i"):
            x = HeavyIrObject("__var~i", self.args)
            x.id = self.id
            y = HeavyIrObject("__varread~i", {"var_id": self.id})

            move_list = []
            for c in [c for cc in self.inlet_connections for c in cc]:
                move_list.append((c, [c.copy(to_object=x)]))
            for c in [c for cc in self.outlet_connections for c in cc]:
                move_list.append((c, [c.copy(from_object=y)]))

            return ({x, y}, move_list)
        elif self.has_inlet_connection_format(["_", "c"]) and self.has_outlet_connection_format(["_", "c"]):
            x = HeavyIrObject("__var", self.args)
        else:
            raise Exception()

        return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlSwitchcase(HeavyObject):

    c_struct = "ControlSwitchase"
    preamble = "cSwichcase"

    @classmethod
    def get_C_def(clazz, obj_type, obj_id):
        return []

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_decl(clazz, obj_type, obj_id, args):
        return [
            f"cSwitchcase_{obj_id}_onMessage(HeavyContextInterface *, void *, int letIn, "
            "const HvMessage *const, void *);"
        ]

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [f"cSwitchcase_{obj_id}_onMessage(_c, NULL, {inlet_index}, m, NULL);"]

    @classmethod
    def get_C_impl(clazz, obj_type, obj_id, on_message_list, obj_class_dict, objects):
        # generate the onMessage implementation
        out_list = [
            f"cSwitchcase_{obj_id}_onMessage(HeavyContextInterface *_c, void *o, int letIn, "
            f"const HvMessage *const m, void *sendMessage) {{"
        ]
        out_list.append("switch (msg_getHash(m, 0)) {")
        cases = objects[obj_id]["args"]["cases"]
        for i, c in enumerate(cases):
            out_list.append("case {0}: {{ // \"{1}\"".format(
                HeavyObject.get_hash_string(c),
                c))
            out_list.extend(
                HeavyObject._get_on_message_list(on_message_list[i], obj_class_dict, objects))
            out_list.append("break;")
            out_list.append("}")
        out_list.append("default: {")
        out_list.extend(
            HeavyObject._get_on_message_list(on_message_list[-1], obj_class_dict, objects))
        out_list.append("break;")
        out_list.append("}")  # end default
        out_list.append("}")  # end switch
        out_list.append("}")  # end function

        return out_list
